"""
Created 2022-05
Moved here 2023-02

@author: Lara
@author: Marc
@author: Wolfgang
"""
from __future__ import annotations

import itertools
import sys
from dbis_tm.TM import Schedule, OperationType
from typing import Union
from graphviz import Digraph


class Serializability:
    """
    I am an interface for checking the serializability of a given schedule (see Schedule class).
    You should not construct me because I am a stateless interface that merely provides static functions.

    Functions:
        conflict_graph_contains_cycle (helper method)
        remove_aborted_tx (helper method)
        is_serializable (checks membership of CSR class)
        build_graphviz_object
    """

    def __init__(self):
        raise TypeError("Cannot create 'Serializability' instances.")

    @classmethod
    def conflict_graph_contains_cycle(
        cls, graph: dict, current: set[int], start: int, visited: set[int]
    ) -> bool:
        """
        Helper method.
        Uses a simple DFS to detect a cycle in `graph`.

        Returns:
            true iff `start` is part of a cycle.
        """
        for dependency in graph[current]:
            if dependency == start:
                return True
            elif dependency not in visited:
                visited.add(dependency)
                has_cycle = cls.conflict_graph_contains_cycle(
                    graph, dependency, start, visited
                )
                if has_cycle:
                    return True
        return False

    @classmethod
    def remove_aborted_tx(cls, schedule: Union[Schedule, str]) -> None:
        """
        Remove aborted transactions in ``schedule`.

        NOTE: This function operates on the passed object!
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        schedule.operations = list(
            filter(lambda op: op.tx_number not in schedule.aborts, schedule.operations)
        )

    @classmethod
    def is_serializable(cls, schedule: Union[Schedule, str]) -> tuple[bool, dict]:
        """
        Check whether `schedule` is serializable, i.e., whether it's conflict graph contains a cycle.

        Returns:
            true iff schedule is serializable
            conflict graph as adjacency-list
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        # graph[1] = {2,3} means that tx 1 is "in conflict" with txs 2 and 3
        graph = {operation.tx_number: set() for operation in schedule.operations}

        # For each operation i in `operation_list` check all operations i + 1 until n
        # for conflicts (i.e., same resource, not same tx, but one of the operations is a write operation)
        for op1, i in zip(schedule.operations, range(len(schedule.operations))):
            for op2 in schedule.operations[i + 1 :]:
                if op1.resource != op2.resource or op1.tx_number == op2.tx_number:
                    continue
                elif (
                    op1.op_type == OperationType.WRITE
                    or op2.op_type == OperationType.WRITE
                ):
                    graph[op1.tx_number].add(op2.tx_number)

        # Check for each tx whether it is part of a cycle in the conflict graph
        graph_has_cycle = any(
            map(lambda n: cls.conflict_graph_contains_cycle(graph, n, n, set()), graph)
        )
        return not graph_has_cycle, graph

    @classmethod
    def build_graphviz_object(cls, graph: dict) -> Digraph:
        """
        Construct Graphviz directed graph from given adjacency list 'graph'.

        Returns:
            Graphviz Digraph object
        """
        dg = Digraph("")
        for t1 in graph:
            for t2 in graph:
                dg.edge(f"t{t1}", f"t{t2}")
        return dg


class Recovery:
    """
    I am an interface for checking the recoverability of a given schedule (see Schedule class).
    You should not construct me because I am a stateless interface that merely provides static functions.

    Functions:
        reads_from (helper method)
        is_recoverable (checks membership of RC class)
        avoids_cascading_aborts (checks membership of ACA class)
        is_strict (checks membership of ST class)
    """

    def __init__(self):
        raise TypeError("Cannot create 'Recovery' instances.")

    @classmethod
    def reads_from(
        cls, schedule: Union[Schedule, str], tx1: int, resource: str, tx2: int
    ) -> Tuple(bool, int):
        """
        Helper method that implements the "reads from" relation:
        We say that any transaction t_i reads any resource x from any transaction t_j if:
            - w_j(x) <_s r_i(x)
            - not (a_j <_s r_i(x))
            - w_j(x) <_s w_k(x) <_s r_i(x) => a_k <_s r_i(x)
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        for op1, i in zip(schedule.operations, range(len(schedule.operations))):
            if not (
                op1.op_type == OperationType.READ
                and op1.tx_number == tx1
                and op1.resource == resource
            ):
                continue
            elif schedule.aborts.get(tx2, sys.maxsize) < op1.index:
                # possible that the same action is done twice in one schedule, have to check both
                return False, op1.index
            for op2 in reversed(schedule.operations[0:i]):
                if op2.op_type == OperationType.WRITE and op2.resource == op1.resource:
                    if schedule.aborts.get(op2.tx_number, sys.maxsize) < op1.index:
                        continue
                    else:
                        # possible that the same action is done twice in one schedule, have to check both
                        if op2.tx_number == tx2:
                            return True, op1.index
                        else:
                            break
        return False, op1.index

    @classmethod
    def is_recoverable(
        cls, schedule: Union[Schedule, str]
    ) -> tuple[bool, set[tuple[int, str, int, bool]]]:
        """
        Check whether `schedule` s is recoverable, i.e., whether the following holds
        for all transaction pairs t_i, t_j with i != j:
            (t_i reads from t_j in s && c_i is in s) => c_j <_s c_i

        Returns:
            true iff schedule is recoverable
            proof if schedule is recoverable
                  else counterexample
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        tx_indices = range(1, schedule.tx_count + 1)

        # A counterexample or proof will contain tuples of the form (t_i, resource, t_j, does t_i read from t_j?).
        # Note that the only way to violate RC is that t_i reads from t_j and t_i commits,
        # BUT t_j commits AFTER t_i or t_j is still ACTIVE (i.e., hasn't committed or aborted yet).
        # Hence, in case of a counterexample we know that c_j <_s c_i
        # DOES NOT hold (t_j has not committed yet or is still active)
        proof = set()
        cex = set()
        for i, r, j in itertools.product(tx_indices, schedule.resources, tx_indices):
            if i == j:
                continue
            reads = cls.reads_from(schedule, i, r, j)[0]
            # Equivalent to ("t_i reads r from t_j" AND "c_i in s") IMPLIES "c_j <_s c_i"
            if not (reads and i in schedule.commits) or (
                schedule.commits.get(j, sys.maxsize) < schedule.commits[i]
            ):
                proof.add((i, r, j, reads))
            else:
                cex.add((i, r, j, reads))
        return (False, cex) if len(cex) != 0 else (True, proof)

    @classmethod
    def avoids_cascading_aborts(
        cls, schedule: Union[Schedule, str]
    ) -> tuple[bool, set[tuple[int, str, int, bool]]]:
        """
        Check whether `schedule` s avoids cascading aborts, i.e., whether the following holds
        for all transaction pairs t_i, t_j with i != j:
            t_i reads from t_j in s => c_j <_s r_i(x)

        Returns:
            true iff schedule avoids cascading aborts
            proof if schedule avoids cascading aborts
                  else counterexample
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        tx_indices = range(1, schedule.tx_count + 1)

        # A counterexample or proof will contain tuples of the form (t_i, resource, t_j, does t_i read from t_j?)
        proof = set()
        cex = set()
        for i, r, j in itertools.product(tx_indices, schedule.resources, tx_indices):
            if i == j:
                continue
            reads, index = cls.reads_from(schedule, i, r, j)
            if reads:
                for op in schedule.operations:
                    if (
                        op.op_type == OperationType.READ
                        and op.tx_number == i
                        and op.resource == r
                        and op.index == index
                        and schedule.commits.get(j, sys.maxsize) >= op.index
                    ):
                        cex.add((i, r, j, reads))
            proof.add((i, r, j, reads))
        return (False, cex) if len(cex) != 0 else (True, proof)

    @classmethod
    def is_strict(
        cls, schedule: Union[Schedule, str]
    ) -> tuple[bool, set[tuple[str, str, bool, bool]]]:
        """
        Check whether `schedule` s is strict, i.e., whether the following holds
        for all transactions t_j and all operations p_i from t_j:
            (w_j(x) <_s p_i(x) && i != j) => (a_j <_s p_i(x) || c_j <_s p_i(x))

        Returns:
            true iff schedule is strict
            proof if schedule is strict
                  else counterexample
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        # A counterexample or proof will contain tuples of the form (w_j(x), p_i(x), a_j <_s p_i(x), c_j <_s p_i(x))
        proof = set()
        cex = set()
        for op1, i in zip(schedule.operations, range(len(schedule.operations))):
            for op2 in reversed(schedule.operations[0:i]):
                if op1.tx_number == op2.tx_number or op1.resource != op2.resource:
                    continue
                elif op2.op_type == OperationType.WRITE:
                    aborted = (
                        schedule.aborts.get(op2.tx_number, sys.maxsize) < op1.index
                    )
                    committed = (
                        schedule.commits.get(op2.tx_number, sys.maxsize) < op1.index
                    )
                    if not (aborted or committed):
                        cex.add((str(op2), str(op1), aborted, committed))
                    else:
                        proof.add((str(op2), str(op1), aborted, committed))
        return (False, cex) if len(cex) != 0 else (True, proof)


class Scheduling:
    """
    I am an interface for checking the whether a schedule (see Schedule class) can be created by a specific scheduler.
    You should not construct me because I am a stateless interface that merely provides static functions.

    Functions:
        is_2PL (checks whether schedule satisfies 2-phase-locking)
        is_C2PL (checks whether schedule satisfies conservative 2-phase-locking)
        is_S2PL (checks whether schedule satisfies strict 2-phase-locking)
        is_SS2PL (checks whether schedule satisfies strong strict 2-phase-locking)
        is_operations_same (Checks whether the two  given schedules do have the same operations.)
    """

    def __init__(self):
        raise TypeError("Cannot create 'Scheduling' instances.")

    @classmethod
    def is_2PL(cls, schedule: Union[Schedule, str]) -> tuple[bool, list[str]]:
        """
         Check whether `schedule` s satisfies 2-phase-locking, i.e., whether the following holds:
            In the first phase locks can only be set.
            In the second phase locks can only be released. Only possible after all locks have been set.

        Returns:
            true iff schedule satisfies 2-phase-locking
            empty list if schedule satisfies 2-phase-locking
                  else counterexample
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        transactions = [[]]  # [[1,2,3,...][#1 ][#2][#3]...]
        locks = []  # all things which have to be locked [[locks of transaction 1][...]]

        # sort by transaction
        for i in schedule.operations:
            if i.tx_number not in transactions[0]:
                transactions[0].extend([i.tx_number])
                transactions.append([i])
                locks.append([])
                if (
                    i.op_type == OperationType.READ_LOCK
                    or i.op_type == OperationType.WRITE_LOCK
                ):
                    locks[len(transactions) - 2].extend(
                        [i.op_type.value[0] + i.resource]
                    )
                    # op_type, tx_number, resource, index)
                    # r, 1, x, an wie vielter stelle
            else:
                index = transactions[0].index(i.tx_number)
                transactions[index + 1].extend([i])
                if (
                    i.op_type == OperationType.READ_LOCK
                    or i.op_type == OperationType.WRITE_LOCK
                ):
                    locks[index].extend([i.op_type.value[0] + i.resource])
        errors = []
        # go through transactions and verify whether they are 2PL
        for i in range(len(transactions[0])):
            locks_set = []
            all_locked = False
            for j in transactions[i + 1]:
                current_op = j.op_type
                representation = j.op_type.value[0] + j.resource

                if locks_set == locks[i]:  # all locks set?
                    all_locked = True

                if (
                    current_op == OperationType.WRITE_LOCK
                    or current_op == OperationType.READ_LOCK
                ):
                    if representation not in locks_set:
                        locks_set.append(representation)
                    else:
                        errors.append(f"--Double lock: {j}")
                elif (
                    current_op == OperationType.WRITE
                    or current_op == OperationType.READ
                ):
                    if representation not in locks_set:
                        errors.append(f"--Not locked before using: {j}")
                elif (
                    current_op == OperationType.WRITE_UNLOCK
                    or current_op == OperationType.READ_UNLOCK
                ):
                    if all_locked:  # all locked?
                        if representation in locks_set:  # already locked?
                            locks_set.remove(representation)
                        else:
                            errors.append(f"--Not locked before unlocking: {j}")
                    else:
                        errors.append(f"--Unlocking before all locks set: {j}")

        is2PL = not errors
        return is2PL, errors

    @classmethod
    def is_C2PL(cls, schedule: Union[Schedule, str]) -> tuple[bool, list[str]]:
        """
        Check whether `schedule` s satisfies conservative 2-phase-locking, i.e., whether the following holds:
            satisfies 2-phase-locking && all locks of a transaction are set before its first operation

        Returns:
            true iff schedule satisfies conservative 2-phase-locking
            empty list if schedule satisfies conservative 2-phase-locking
                  else counterexample
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        res = cls.is_2PL(schedule)
        if not res[0]:
            return False, res[1]

        for i in range(1, schedule.tx_count + 1):
            tx_ops = list(filter(lambda op: op.tx_number == i, schedule.operations))
            index_last_lock = next(
                (
                    tx_ops.index(op)
                    for op in reversed(tx_ops)
                    if op.op_type in [OperationType.READ_LOCK, OperationType.WRITE_LOCK]
                ),
                sys.maxsize,
            )
            index_first_op = next(
                (
                    tx_ops.index(op)
                    for op in tx_ops
                    if op.op_type in [OperationType.WRITE, OperationType.READ]
                ),
                -sys.maxsize,
            )
            if index_last_lock >= index_first_op:
                lock = next(
                    (
                        op
                        for op in tx_ops
                        if op.op_type
                        in [OperationType.READ_LOCK, OperationType.WRITE_LOCK]
                    )
                )
                return False, [f"Lock {lock} was acquired after first r/w operation"]
        return True, []

    @classmethod
    def is_S2PL(cls, schedule: Union[Schedule, str]) -> tuple[bool, list[str]]:
        """
        Check whether `schedule` s satisfies strict 2-phase-locking, i.e., whether the following holds:
            satisfies 2-phase-locking && all write locks of a transaction are held until its last operation

        Returns:
            true iff schedule satisfies strict 2-phase-locking
            empty list if schedule satisfies strict 2-phase-locking
                  else counterexample
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        res = cls.is_2PL(schedule)
        if not res[0]:
            return False, res[1]

        for i in range(1, schedule.tx_count + 1):
            tx_ops = list(filter(lambda op: op.tx_number == i, schedule.operations))
            index_first_unlock = next(
                (
                    tx_ops.index(op)
                    for op in tx_ops
                    if op.op_type == OperationType.WRITE_UNLOCK
                ),
                sys.maxsize,
            )
            index_last_op = next(
                (
                    tx_ops.index(op)
                    for op in reversed(tx_ops)
                    if op.op_type in [OperationType.WRITE, OperationType.READ]
                ),
                -sys.maxsize,
            )
            if index_first_unlock <= index_last_op:
                unlock = next(
                    (op for op in tx_ops if op.op_type == OperationType.WRITE_UNLOCK)
                )
                return False, [f"Unlock {unlock} was done before last r/w operation"]
        return True, []

    @classmethod
    def is_SS2PL(cls, schedule: Union[Schedule, str]) -> tuple[bool, list[str]]:
        """
        Check whether `schedule` s satisfies strong strict 2-phase-locking, i.e., whether the following holds:
            satisfies 2-phase-locking && all read/write locks of a transaction are held until its last operation

        Returns:
            true iff schedule satisfies strong strict 2-phase-locking
            empty list if schedule satisfies strong strict 2-phase-locking
                  else counterexample
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        res = cls.is_2PL(schedule)
        if not res[0]:
            return False, res[1]

        for i in range(1, schedule.tx_count + 1):
            tx_ops = list(filter(lambda op: op.tx_number == i, schedule.operations))
            index_first_unlock = next(
                (
                    tx_ops.index(op)
                    for op in tx_ops
                    if op.op_type
                    in [OperationType.READ_UNLOCK, OperationType.WRITE_UNLOCK]
                ),
                sys.maxsize,
            )
            index_last_op = next(
                (
                    tx_ops.index(op)
                    for op in reversed(tx_ops)
                    if op.op_type in [OperationType.WRITE, OperationType.READ]
                ),
                -sys.maxsize,
            )
            if index_first_unlock <= index_last_op:
                unlock = next(
                    (
                        op
                        for op in tx_ops
                        if op.op_type
                        in [OperationType.READ_UNLOCK, OperationType.WRITE_UNLOCK]
                    )
                )
                return False, [f"Unlock {unlock} was done before last r/w operation"]
        return True, []



""" 
Created on 2022-07-16

@author: wf
"""
from dbis_tm.TMCheck import (
    ConflictSetScorer,
    ConflictSerializationScorer,
    RecoveryScorer,
    ScheduleScorer,
)
from dbis_tm.TM import ConflictGraph, ConflictGraphNode
from tests.scheduletest import ScheduleTest


class TestTMScorer(ScheduleTest):
    """
    tests for scorers of
    for https://git.rwth-aachen.de/i5/teaching/dbis-digi-2022/-/issues/35
    """

    def testConflictSetScorer(self):
        """
        test the conflict set scorer
        """
        examples = [
            {
                "result1": {("w_2(x)", "r_1(x)"), ("w_1(z)", "w_2(z)")},
                "result2": {
                    ("r_2(x)", "w_3(x)"),
                    ("w_1(y)", "r_2(y)"),
                    ("w_1(y)", "w_2(y)"),
                    ("w_1(y)", "w_3(y)"),
                    ("w_3(z)", "w_1(z)"),
                    ("w_3(z)", "w_2(z)"),
                    ("r_2(y)", "w_3(y)"),
                    ("w_2(y)", "w_3(y)"),
                    ("w_2(y)", "r_1(y)"),
                    ("w_1(z)", "w_2(z)"),
                    ("w_3(y)", "r_1(y)"),
                    ("w_3(y)", "w_2(y)"),
                    ("r_1(y)", "w_2(y)"),
                },
                "expected": 4.0,
            }
        ]
        solution1 = "w2(x) r2(x) w1(z) r1(x) w3(y) a3 w2(z) c1 c2 "  # {("w_2(x)", "r_1(x)"), ("w_1(z)", "w_2(z)")}
        solution2 = "r2(x) w3(x) w1(y) r2(y) w2(y) w3(y) w3(z) w1(z) w2(z) r1(y) w2(y)"
        # {("r_2(x)", "w_3(x)"),- ("w_1(y)", "r_2(y)"), ("w_1(y)", "w_2(y)"), ("w_1(y)", "w_3(y)"),- ("w_3(z)", "w_1(z)"), ("w_3(z)", "w_2(z)"), ("r_2(y)", "w_3(y)"),
        #  ("w_2(y)", "w_3(y)"), ("w_2(y)", "r_1(y)"), ("w_1(z)", "w_2(z)"), ("w_3(y)", "r_1(y)"), ("w_3(y)", "w_2(y)"), ("r_1(y)", "w_2(y)")}
        max_points = 4
        debug = True
        for i, example in enumerate(examples):
            result1 = example["result1"]
            result2 = example["result2"]
            expected = example["expected"]
            css = ConflictSetScorer()
            score = css.score_conflictSet(
                result1, result2, solution1, solution2, max_points
            )
            if debug:
                print(f"{i+1} score: {score} expected: {expected}")
            self.assertEqual(expected, score)

    def testConflictSerializationScorer(self):
        """
        test conflict serialization scoring
        """
        schedule1 = "r1(x) w2(x) w3(x) w2(y) w3(y) r1(y) w3(z) r1(z) w2(z) c1 c2 c3"
        schedule2 = "w1(x) w3(x) w3(y) w1(y)  r2(z) c2 c1 c3"
        t1 = ConflictGraphNode(1)
        t2 = ConflictGraphNode(2)
        t3 = ConflictGraphNode(3)

        graph1 = ConflictGraph()
        graph1.add_edge(t1, t2)
        graph1.add_edge(t1, t3)
        graph1.add_edge(t2, t1)
        graph1.add_edge(t2, t3)
        graph1.add_edge(t3, t1)
        graph1.add_edge(t3, t2)

        graph2 = ConflictGraph()
        graph2.add_edge(t1, t3)
        graph2.add_edge(t3, t1)

        conf_g3 = ConflictGraph()
        conf_g3.add_edge(t1, t2)
        examples = [
            {
                "name": "s3",
                "result": graph1,
                "serializable": False,
                "schedule": schedule1,
                "expected": 1.5,
            },
            {
                "name": "s3",
                "result": conf_g3,
                "serializable": False,
                "schedule": schedule1,
                "expected": 0.5,
            },
            {
                "name": "s4",
                "result": graph2,
                "schedule": schedule2,
                "serializable": False,
                "expected": 1.5,
            },
        ]
        debug = False
        for i, example in enumerate(examples):
            name = example["name"]
            result = example["result"]
            schedule = example["schedule"]
            expected = example["expected"]
            serializableResult = example["serializable"]

            css = ConflictSerializationScorer()
            score = css.score_conflictSerialization(
                name, result, serializableResult, schedule, 1.5
            )
            print(score)
            if debug:
                print(f"{i+1} score: {score} expected: {expected}")
            self.assertEqual(expected, score)

    def testRecoveryScorer(self):
        """
        test the recovery scorer
        """
        # subtask2.setSolution(([s2_is_rc, s2_is_aca, s2_is_st], [s2_proof_rc, s2_proof_aca, s2_proof_st]))

        s1 = "w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1"
        s3 = "w_2(y) r_1(z) c_2 w_1(y) r_1(y) w_1(z) c_1"

        examples = [
            # No points -no result specified
            {
                "name": "s1",
                "schedule": s1,
                "expected": 0,
                "result": (
                    (None, None, None),  # is_rc  # is_aca  # is_st
                    (
                        # proof_rc
                        {},
                        # proof_aca
                        {},
                        # proof_st
                        {},
                    ),
                ),
            },
            # No points, is in rc but no proof given
            {
                "name": "s1",
                "schedule": s1,
                "expected": 0,
                "result": (
                    (True, False, False),  # is_rc  # is_aca  # is_st
                    (
                        # proof_rc
                        {},
                        # proof_aca
                        {},
                        # proof_st
                        {},
                    ),
                ),
            },
            # No points, is in rc but no proof given
            {
                "name": "s1",
                "schedule": s1,
                "expected": 0,
                "result": (
                    (True, True, True),  # is_rc  # is_aca  # is_st
                    (
                        # proof_rc
                        {},
                        # proof_aca
                        {},
                        # proof_st
                        {},
                    ),
                ),
            },
            # correct answer and proof
            {
                "name": "s1",
                "schedule": s1,
                "expected": 2,
                "result": (
                    (True, False, False),  # is_rc  # is_aca  # is_st
                    (
                        # proof_rc
                        {
                            (1, "y", 2, True),
                            (2, "y", 1, False),
                            (1, "z", 2, False),
                            (2, "z", 1, False),
                        },
                        # proof_aca
                        {(1, "y", 2, True)},
                        # proof_st
                        {},
                    ),
                ),
            },
            {
                "name": "s1",
                "schedule": s1,
                "expected": 0,
                "result": (
                    (True, False, False),  # is_rc  # is_aca  # is_st
                    (
                        # proof_rc
                        {},
                        # proof_aca
                        {},
                        # proof_st
                        {},
                    ),
                ),
            },
            {
                "name": "s3",
                "schedule": s3,
                "expected": 2,
                "result": (
                    (True, True, True),  # is_rc  # is_aca  # is_st
                    (
                        # proof_rc
                        {
                            (1, "y", 2, False),
                            (2, "y", 1, False),
                            (1, "z", 2, False),
                            (2, "z", 1, False),
                        },
                        # proof_aca
                        {
                            (1, "y", 2, False),
                            (2, "y", 1, False),
                            (1, "z", 2, False),
                            (2, "z", 1, False),
                        },
                        # proof_st
                        {
                            ("w_2(y)", "w_1(y)", False, True),
                            ("w_2(y)", "r_1(y)", False, True),
                        },
                    ),
                ),
            },
        ]
        debug = False
        for i, example in enumerate(examples):
            name = example["name"]
            result = example["result"]
            schedule = example["schedule"]
            expected = example["expected"]
            rs = RecoveryScorer()
            score = rs.score_recovery(name, schedule, result, 2)

            if debug:
                print(f"{i+1} score: {score} expected: {expected}")
            self.assertEqual(expected, score)

    def testScheduleScorer(self):
        """
        test the Schedule scorer
        """
        examples = self.getScheduleExamples()
        debug = False
        for i, example in enumerate(examples):
            index = example["index"]
            schedule = example["schedule"]
            result = example["result"]
            check = example["check"]
            sScorer = ScheduleScorer()
            score = sScorer.getScore(schedule, result, check, max_points=1)
            if debug:
                print(f"{i+1}: score={score}")
            self.assertEqual(1, score)

    # def grade_recovery(self,schedule,isClassList,proofList,max_score):
    #     rs=RecoveryScorer()
    #     name="?"
    #     result=(tuple(isClassList),tuple(proofList))
    #     score=rs.score_proof(name,schedule,result, max_score)
    #     return score

    # def testGradeRecovery(self):
    #     """
    #     test Grade Recovery
    #     """
    #     # No points
    #     rc_proof = {}
    #     is_rc = None
    #     aca_proof = {}
    #     is_aca = None
    #     st_proof = {}
    #     is_st = None
    #     self.assertEqual(0,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # No points, is in rc but no proof given
    #     is_rc = True
    #     is_aca = None
    #     is_st = None
    #     self.assertEqual(0,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # No points, is in rc but no proof given
    #     is_rc = True
    #     is_aca = True
    #     is_st = True
    #     self.assertEqual(0,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # Gets 0 because is_aca was not shown although is_st == False is correct
    #     is_rc = True
    #     is_aca = False
    #     is_st = False
    #     self.assertEqual(0,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # Gets 0 because is_st is correct but is_aca was not shown
    #     is_rc = True
    #     is_aca = True
    #     is_st = False
    #     self.assertEqual(0,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # Gets 2 / 3 because is_aca (cex given)
    #     is_rc = None
    #     is_aca = False
    #     aca_proof = {(1, 'y', 2, True)}
    #     is_st = None
    #     self.assertEqual(2 / 3,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # Gets 2 * (2 / 3) because is_aca (cex given) and is_st is correct
    #     is_rc = True
    #     is_aca = False
    #     aca_proof = {(1, 'y', 2, True)}
    #     is_st = False
    #     self.assertEqual(4 / 3,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # Gets 0 points because is_aca (cex given) but bool is not set
    #     is_rc = None
    #     is_aca = None
    #     aca_proof = {(1, 'y', 2, True)}
    #     is_st = None
    #     self.assertEqual(0,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # Gets 2 / 3 points because is_aca (cex given) and bool is set
    #     is_rc = None
    #     is_aca = False
    #     aca_proof = {(1, 'y', 2, True)}
    #     is_st = None
    #     self.assertEqual(2 / 3,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # Gets 0 points because is_aca == False was not shown although is_st == False is correct
    #     is_rc = None
    #     is_aca = None
    #     aca_proof = {}
    #     is_st = False
    #     self.assertEqual(0,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))

    #     # Gets 2 points (correct solution)
    #     is_rc = True
    #     rc_proof = {(1, 'z', 2, False), (2, 'y', 1, False), (1, 'y', 2, True), (2, 'z', 1, False)}
    #     is_aca = False
    #     aca_proof = {(1, 'y', 2, True)}
    #     is_st = False
    #     st_proof = {}
    #     self.assertEqual(2,
    #                      self.grade_recovery("w_2(y) r_1(z) r_2(y) r_1(y) c_2 w_1(y) c_1", [is_rc, is_aca, is_st],
    #                                             [rc_proof, aca_proof, st_proof], 2))



from Generate_Tasks import get_tasks, Creating
from dbis_tm.Generate import generate
from dbis_tm.TM import Schedule
from dbis_tm.TMSolver import Scheduling, Serializability
from dbis_tm.Solution_generator import Perform_scheduling, Perform_conflictgraph
import random

# random.seed(10)
helper = get_tasks()
print(helper)
print(
    helper[0][0],
    "\n",
    helper[0][1],
    "\n",
    "\n",
    helper[1][0],
    "\n",
    helper[1][1],
    "\n",
    helper[1][2],
    "\n",
    "\n",
    helper[2],
    "\n",
)


# for exams:
# recovery possible to
recovery_list = []
for i in range(10):
    recovery = Creating()
    tasks = recovery.create_recovery()
    recovery_list.append(tasks.subtask1)
    recovery_list.append(tasks.subtask2)
    recovery_list.append(tasks.subtask3)
# or
recovery_list2 = [[], [], [], []]
resources = [["a", "b", "c"], ["x", "y", "z"], ["u", "r", "d"], ["f", "a", "x"]]
for i in range(10):
    none = generate(3, random.choice(resources), recovery="n")[0]
    recovery = generate(3, random.choice(resources), recovery="r")[0]
    avoid = generate(3, random.choice(resources), recovery="a")[0]
    strict = generate(3, random.choice(resources), recovery="s")[0]
    recovery_list2[0].append(Schedule.parse_string(none)[0])
    recovery_list2[1].append(Schedule.parse_string(recovery)[0])
    recovery_list2[2].append(Schedule.parse_string(avoid)[0])
    recovery_list2[3].append(Schedule.parse_string(strict)[0])
# print(recovery_list2[0],'\n','\n',recovery_list2[1],'\n','\n',recovery_list2[2],'\n','\n',recovery_list2[3])
# scheduling
schedules = [[], [], [], [], []]
for i in range(10):
    schedule = generate(3, random.choice(resources), deadlock=False)
    deadlock = generate(3, random.choice(resources), deadlock=True)
    schedules[3].append(Schedule.parse_string(deadlock[0])[0])
    if schedule[1] != "":
        continue
    else:
        schedule = schedule[0]
        s2pl = Perform_scheduling.perform_S2PL(schedule)[0]
        ss2pl = Perform_scheduling.perform_SS2PL(schedule)[0]
        c2pl = Perform_scheduling.perform_C2PL(schedule)
        schedules[0].append(Schedule.parse_string(s2pl)[0])
        schedules[1].append(Schedule.parse_string(ss2pl)[0])
        schedules[2].append(Schedule.parse_string(c2pl)[0])
        schedules[4].append(Schedule.parse_string(schedule)[0])
# print(schedules[0],'\n','\n',schedules[1],'\n','\n',schedules[2],'\n','\n', schedules[3],'\n','\n', schedules[4])

# conflicts
conflicts = [[], [], [], []]  # serilaziable, not-seri,conf-graphs(dict), conflict-sets
for i in range(10):
    schedule = generate(3, random.choice(resources))[0]
    seri = Serializability.is_serializable(schedule)
    if seri[0]:
        conflicts[0].append(Schedule.parse_string(schedule)[0])
    else:
        conflicts[1].append(Schedule.parse_string(schedule)[0])
    conflicts[2].append(seri[1])
    conflicts[3].append(Perform_conflictgraph.compute_conflictgraph(seri[1]))
# print(conflicts[0],'\n','\n',conflicts[1],'\n','\n',conflicts[2],'\n','\n', conflicts[3],'\n','\n', )



from dbis_tm.TM import Schedule, OperationType, Operation
from dbis_tm.TMSolver import Recovery
import random
import copy
from typing import Union
from dbis_tm.Solution_generator import predict_deadlock


def generate(
    transactions: int, resources: list[str], deadlock=None, recovery=None
) -> tuple[Schedule, str]:
    schedule = generate_schedule(transactions, resources, deadlock, recovery)
    schedule_test = copy.deepcopy(schedule)
    if deadlock == False:
        for i in range(10):
            if not predict_deadlock(schedule_test):
                break
            schedule = generate_schedule(transactions, resources, deadlock)
            schedule_test = copy.deepcopy(schedule)
        if predict_deadlock(schedule_test):
            return schedule, "Please try again."
    return schedule, ""


def generate_schedule(
    transactions: int, resources: list[str], deadlock=None, recovery=None
) -> Schedule:
    case = ""
    if bool(deadlock) and bool(recovery):
        raise ValueError("Not allowed to choose deadlock and recovery.")
    elif deadlock in [True, False]:
        case = "deadlock"
    elif recovery in ["r", "a", "s", "n"]:
        case = "recovery"
    # check_res
    check_res = [0] * len(resources)
    # initiate schedule parts
    operations = []
    aborts = dict()
    commits = dict()
    # initiate lists for random choice, and length
    index = random.randrange(transactions * 3, transactions * 5 + 1)
    operation_ch = [OperationType.READ, OperationType.WRITE]
    conclude_choice = [1, 0, 0]

    transl = list([0] for i in range(transactions))
    transl.insert(0, list(range(1, transactions + 1)))

    # deadlock
    locks = [[] for i in range(len(resources))]  # [[[op,res,trans]#res a...]...]
    circle = [
        [] for i in range(transactions)
    ]  # [[1 waiting for trans]...], add resources, mabye with dicts
    deadlock_occured = False
    # recovery
    list_write = []  # list of last write action on resource
    reads = [[] for i in range(transactions)]  # [[1 reads from trans]...]
    not_next = False
    i = 1

    while index > 0:
        trans = 0
        if index == len(transl[0]):
            for u in transl[0]:
                if sum(transl[u]) != 0:
                    transl[u] = [4]
                else:
                    index += 1
                    trans = copy.deepcopy(u)
        if trans == 0:
            # choose random transaction, operation and resource
            trans = random.choice(transl[0])

        if transl[trans][0] == 0 or (
            len(transl[0]) == 1 and index > 1
        ):  # if none performed, if not the last action, but only one operation remains
            conclude = False
        elif (
            transl[trans][0] >= 4
        ):  # if all performed, must be in this position, so one transaction in not forced to performed till the end by more then 4 times
            conclude = True
        else:  # else (between 1-3 actions)
            conclude = bool(random.choice(conclude_choice))

        if case == "deadlock":
            (
                transl,
                locks,
                circle,
                deadlock_occured,
                commits,
                aborts,
                op,
                trans,
                res,
            ) = generate_deadlock(
                transactions,
                resources,
                deadlock,
                operation_ch,
                index,
                conclude,
                i,
                transl,
                locks,
                circle,
                deadlock_occured,
                trans,
                commits,
                aborts,
            )

        elif case == "recovery":
            (
                index,
                list_write,
                reads,
                not_next,
                transl,
                commits,
                aborts,
                op,
                trans,
                res,
            ) = generate_recovery(
                transactions,
                resources,
                operation_ch,
                recovery,
                index,
                i,
                list_write,
                reads,
                not_next,
                transl,
                trans,
                conclude,
                commits,
                aborts,
                operations,
            )

        else:
            if conclude:  # add an abort/commit
                if bool(random.choice([0, 1])):
                    aborts[trans] = i
                else:
                    commits[trans] = i
                transl[0].remove(trans)
                trans = trans * (-1)

            else:  # add an action
                op = random.choice(operation_ch)
                res = random.choice(resources)
                transl[trans][0] += 1

        if trans == 0:
            continue
        elif trans < 0:  # conclude
            trans = trans * (-1)
        else:
            check_res[resources.index(res)] = 1
            gen_op = Operation(op, trans, res, i)
            operations.append(gen_op)

        index -= 1
        i += 1
    resources_clean = [
        name for name in resources if check_res[resources.index(name)] == 1
    ]
    schedule = Schedule(operations, resources_clean, transactions, aborts, commits)
    return schedule


def generate_recovery(
    transactions: int,
    resources: list[str],
    operation_ch: list,
    recovery: str,
    index: int,
    i: int,
    list_write: list,
    reads: list,
    not_next: bool,
    transl: list,
    trans: int,
    conclude: bool,
    commits,
    aborts,
    operations,
) -> tuple[int, list, list, bool, list, dict, dict, OperationType, int, str]:
    prepared = bool([read for x in reads for read in x])
    if (
        len(transl[0]) == 2
        and not not_next
        and (recovery == "a" or (recovery in ["n", "r"] and not prepared))
    ):
        conclude = False

    if conclude:  # add an commit
        if (
            bool(random.choice([0, 0, 1]))
            and not (recovery in ["n", "r"] and len(transl[0]) <= 2 and not not_next)
            and not (
                recovery == "r" and [read for x in reads for read in x if read == trans]
            )
        ):
            aborts[trans] = i
            reads = [[d for d in r if d != trans] for r in reads]  #
            prepared = bool([read for x in reads for read in x])
            if list_write != [
                write for write in list_write if write.tx_number != trans
            ]:
                operations2 = copy.deepcopy(operations)
                operations2.reverse()
                list_write = []
                for r in resources:
                    # in commits have to be included but are not allowed to be wirtten into the list
                    list_res_op = list(
                        filter(
                            lambda op: op.resource == r
                            and op.tx_number != trans
                            and op.tx_number
                            and op.tx_number not in aborts
                            and op.op_type.value == "w",
                            operations2,
                        )
                    )
                    if list_res_op and list_res_op[0].tx_number not in commits:
                        list_write.append(list_res_op[0])
            if recovery in ["n", "r"]:
                if not prepared and len(transl[0]) <= index and not not_next:
                    if not [write for write in list_write if write.tx_number != trans]:
                        index = len(transl[0]) + 2
                    else:
                        index = len(transl[0]) + 1

        else:
            if recovery == "r":
                if reads[trans - 1]:  # if reading from someone who not commited/aborded
                    if [t for t in transl[0] if reads[t - 1] == []]:  #
                        new_trans = [
                            q + 1
                            for q, val in enumerate(reads)
                            if val == [] == reads[q] and q + 1 in transl[0]
                        ]
                        trans = random.choice(new_trans)
                    else:
                        return (
                            index,
                            list_write,
                            reads,
                            not_next,
                            transl,
                            commits,
                            aborts,
                            operation_ch[0],
                            0,
                            resources[0],
                        )
                if [r for read in reads for r in read if r == trans]:
                    not_next = True

            elif recovery == "n" and prepared and not not_next:
                # check whether critical
                if len(transl[0]) == 2:  # critical
                    if not reads[trans - 1]:
                        trans = random.choice([x for x in transl[0] if x != trans])
                    not_next = True
                elif reads[trans - 1]:
                    not_next = True
            commits[trans] = i
            list_write = [write for write in list_write if write.tx_number != trans]
        reads = [[d for d in r if d != trans] for r in reads]
        reads[trans - 1] = []

        transl[0].remove(trans)
        return (
            index,
            list_write,
            reads,
            not_next,
            transl,
            commits,
            aborts,
            operation_ch[0],
            int(-trans),
            resources[0],
        )

    else:  # add an action
        op = random.choice(operation_ch)
        res = random.choice(resources)
        if recovery == "n":
            if (
                index <= len(transl[0]) + 1 and not not_next and not prepared
            ):  # no reads from relation yet
                if list_write:  # perform read
                    if index < len(transl[0]) + 1:
                        index = len(transl[0]) + 1
                    action = random.choice(list_write)
                    op = OperationType.READ
                    res = action.resource
                    trans = random.choice(
                        [x for x in transl[0] if x != action.tx_number]
                    )
                    reads[trans - 1] = list(set(reads[trans - 1] + [action.tx_number]))
                    transl[trans][0] += 1
                else:  # perform write
                    if index < len(transl[0]) + 1:
                        index = len(transl[0]) + 2
                    op = OperationType.WRITE
                    transl[trans][0] += 1
                    list_write = [r for r in list_write if r.resource != res]
                    list_write.append(Operation(op, trans, res, i))
            else:
                if op.value == "r":
                    write = list(
                        filter(
                            lambda op: op.resource == res and op.tx_number != trans,
                            list_write,
                        )
                    )
                    if write:
                        h_write = [s.tx_number for s in write]
                        reads[trans - 1] = list(set(reads[trans - 1] + h_write))
                else:
                    list_write = [r for r in list_write if r.resource != res]
                    list_write.append(Operation(op, trans, res, i))
                transl[trans][0] += 1

        elif recovery == "r":
            write = list(
                filter(
                    lambda op: op.resource == res and op.tx_number != trans, list_write
                )
            )
            if index == len(transl[0]) + 1 and not prepared and not not_next:
                if not write:
                    if list_write:  # check for other possibilities
                        operation_ch = random.choice(list_write)
                        res = operation_ch.resource
                        op = OperationType.READ
                        trans = random.choice(
                            [x for x in transl[0] if x != operation_ch.tx_number]
                        )
                        write_new = list(
                            filter(
                                lambda op: op.resource == res and op.tx_number != trans,
                                list_write,
                            )
                        )
                        h_write = [s.tx_number for s in write_new]
                        reads[trans - 1] = list(set(reads[trans - 1] + h_write))
                        # prepared = True
                        transl[trans][0] += 1
                    else:
                        op = OperationType.WRITE
                        index += 1
                        transl[trans][0] += 1
                        list_write = [r for r in list_write if r.resource != res]
                        list_write.append(Operation(op, trans, res, i))
                else:
                    op = OperationType.READ
                    h_write = [s.tx_number for s in write]
                    reads[trans - 1] = list(set(reads[trans - 1] + h_write))
                    transl[trans][0] += 1
                    # prepared = True
                # make sure one action is performed
            elif (
                op.value == "r"
            ):  # have to check only the latest write on the current res
                h_write = [s.tx_number for s in write]
                read = list(
                    filter(
                        lambda op: reads.index(op) + 1 in h_write and trans in op, reads
                    )
                )
                reads_temp = copy.deepcopy(reads)
                reads_temp[trans - 1] = list(set(h_write + reads[trans - 1]))
                if not check_circle(reads_temp):
                    # test if any of the write actions is already reading from
                    transl[trans][0] += 1
                    if write:  # mark read transaction
                        reads[trans - 1] = list(set(h_write + reads[trans - 1]))
                        prepared = True
                else:
                    return (
                        index,
                        list_write,
                        reads,
                        not_next,
                        transl,
                        commits,
                        aborts,
                        op,
                        0,
                        res,
                    )
            else:
                transl[trans][0] += 1
                list_write = [r for r in list_write if r.resource != res]
                list_write.append(Operation(op, trans, res, i))

        elif recovery == "a":  # r has to come before commit but a
            # write action has to be performed before commiting
            write = list(
                filter(
                    lambda op: op.resource == res and op.tx_number != trans, list_write
                )
            )
            if index == len(transl[0]) + 1 and not not_next:
                if write:
                    op = OperationType.WRITE
                    not_next = True
                else:  # force action
                    if list_write:  # check for other possibilities
                        operation_ch = random.choice(list_write)
                        res = operation_ch.resource
                        op = operation_ch.op_type
                        trans = random.choice(
                            [x for x in transl[0] if x != operation_ch.tx_number]
                        )
                        not_next = True
                    else:  # perform write and another write next time
                        index += 1
                        op = OperationType.WRITE
                transl[trans][0] += 1
                list_write = [r for r in list_write if r.resource != res]
                list_write.append(Operation(op, trans, res, i))

            if (
                op.value == "r"
            ):  # test if the action we are reading from is already commited
                if not write:
                    transl[trans][0] += 1
                elif not not_next:  # perform action as write action
                    op = OperationType.WRITE
                    list_write = [r for r in list_write if r.resource != res]
                    list_write.append(Operation(op, trans, res, i))
                    not_next = True
                else:
                    return (
                        index,
                        list_write,
                        reads,
                        not_next,
                        transl,
                        commits,
                        aborts,
                        op,
                        0,
                        res,
                    )
            else:
                transl[trans][0] += 1
                list_write = [r for r in list_write if r.resource != res]
                list_write.append(Operation(op, trans, res, i))

        elif recovery == "s":
            write = list(
                filter(
                    lambda op: op.resource == res and op.tx_number != trans, list_write
                )
            )
            if not write:
                transl[trans][0] += 1
                if op.value == "w":
                    list_write = [r for r in list_write if r.resource != res]
                    list_write.append(Operation(op, trans, res, i))
            else:
                return (
                    index,
                    list_write,
                    reads,
                    not_next,
                    transl,
                    commits,
                    aborts,
                    op,
                    0,
                    res,
                )
    return index, list_write, reads, not_next, transl, commits, aborts, op, trans, res


def generate_deadlock(
    transactions: int,
    resources: list,
    deadlock,
    operation_ch: list,
    index: int,
    conclude: bool,
    i: int,
    transl: list,
    locks: list,
    circle: list,
    deadlock_occured: bool,
    trans: int,
    commits,
    aborts,
) -> tuple[list, list, list, bool, dict, dict, OperationType, int, str]:
    if (
        deadlock and not deadlock_occured and len(transl[0]) == 2
    ):  # no deadlock yet, must occur, needs al least 2 trans
        conclude = False

    if conclude:  # add an commit (no abort here)
        transl[0].remove(trans)
        commits[trans] = i
        if not circle[trans - 1]:  # if waiting no actions can be performed
            for e in locks:  # clear locks
                ind = locks.index(e)
                locks[ind] = list(filter(lambda op: op[2] != trans, e))

            circle[trans - 1] = []
            for t in circle:  # clear circle
                if trans in t:
                    t.remove(trans)
        op = operation_ch[0]
        trans = int(-trans)
        res = resources[0]

    else:  # add an action
        op = random.choice(operation_ch)
        res = random.choice(resources)
        len_locks = 0  # needed cause locks is an multidimensional list
        trans_first_locks = []  # list of all transactions performing !first! locks
        list_filtered = []
        for h in locks:
            if h:
                trans_first_locks.append(h[0][2])
                list_filtered.append(h[0])
        trans_first_locks = list(set(trans_first_locks))
        for k in locks:
            len_locks += len(k)

        if deadlock:
            if deadlock_occured:
                transl[trans][0] += 1  # 1
                ind_res = resources.index(res)
                locks[ind_res].append([op, res, trans])
            else:  # must be altered for random generation
                # step1 first locks
                if len(trans_first_locks) < len(resources) and len(
                    trans_first_locks
                ) < len(transl[0]):
                    if len_locks != 0:
                        if list(
                            filter(
                                lambda op: op[2] == trans and op[2 in transl[0]],
                                list_filtered,
                            )
                        ):
                            # transaction does not hold a first lock
                            trans_ch = list(
                                filter(
                                    lambda op: op not in trans_first_locks, transl[0]
                                )
                            )
                            trans = random.choice(trans_ch)

                        if list(filter(lambda op: op[1] == res, list_filtered)):
                            # resource is not already locked
                            res_ch = []
                            for u in locks:
                                if not u:
                                    res_ch.append(resources[locks.index(u)])
                            res = random.choice(res_ch)
                elif len(transl[0]) == len(trans_first_locks) or len(
                    trans_first_locks
                ) == len(resources):
                    # perform waiting
                    waiting = []  # waiting on
                    for k in circle:
                        if k:
                            for w in k:
                                waiting.append(w)
                    waiting = list(set(waiting))
                    if not waiting:
                        list_filtered = list(
                            filter(lambda op: op[2] != trans, list_filtered)
                        )
                        action = random.choice(list_filtered)
                        if action[0].value == "r":
                            op = OperationType.WRITE
                        else:
                            op = random.choice(operation_ch)
                        res = action[1]
                        circle[trans - 1].append(action[2])
                    else:
                        if circle[
                            trans - 1
                        ]:  # check whether trans has already performed
                            # get different trans
                            counter = 1
                            for p in circle:
                                if not p and counter in transl[0]:
                                    trans = counter
                                    break
                                counter += 1
                        list_filtered = list(
                            filter(
                                lambda op: op[2] not in waiting and op[2] != trans,
                                list_filtered,
                            )
                        )
                        action = random.choice(list_filtered)
                        if action[0].value == "r":
                            op = OperationType.WRITE
                        else:
                            op = random.choice(operation_ch)
                        res = action[1]
                        circle[trans - 1].append(action[2])
                if check_circle(circle):
                    deadlock_occured = True
                transl[trans][0] += 1
                ind_res = resources.index(res)
                locks[ind_res].append([op, res, trans])
        else:
            # checking for deadlock:
            problem = []  # check for potential problem
            i_cur_res = resources.index(res)
            if locks[i_cur_res]:
                if locks[i_cur_res][0][2] != trans and (
                    locks[i_cur_res][0][0] != op or locks[i_cur_res][0][0].value == "w"
                ):
                    problem.append(
                        locks[i_cur_res][0][2]
                    )  # check whether other trans and whether not both read
            problem = list(set(problem + circle[trans - 1]))
            circle_temp = copy.deepcopy(circle)
            circle_temp[trans - 1] = problem

            if check_circle(circle_temp):
                trans = 0
            else:
                transl[trans][0] += 1
                ind_res = resources.index(res)
                locks[ind_res].append([op, res, trans])
                circle = circle_temp
    return transl, locks, circle, deadlock_occured, commits, aborts, op, trans, res
    # problem deadlock after one has concluded but was still waiting
    # could just check wether it worked, otherwise try again


def check_circle(transactions: list()) -> bool:
    trans = len(transactions)
    ignore = []
    help = [o + 1 for o in range(len(transactions))]
    i = 0
    while i < len(transactions):
        if (
            not list(filter(lambda op: op not in ignore, transactions[i]))
            and i + 1 not in ignore
        ):
            ignore.append(i + 1)
            i = -1
        i += 1

    ignore = list(set(ignore))
    if len(ignore) == len(transactions):
        return False
    return True



from dbis_tm.TM import Schedule, ConflictGraph
from dbis_tm.TMSolver import Recovery, Scheduling
from dbis_tm.Generate import generate
from dbis_tm.Solution_generator import (
    predict_deadlock,
    Perform_scheduling,
    Perform_conflictgraph,
)
from unittest import TestCase


class TestTMPerformance(TestCase):
    def test_generate_recovery(self):
        test_number = 1000
        for i in range(test_number):
            none = generate(3, ["a", "b", "c"], recovery="n")[0]
            self.assertFalse(
                Recovery.is_recoverable(none)[0], Schedule.parse_string(none)
            )
            self.assertFalse(
                Recovery.avoids_cascading_aborts(none)[0], Schedule.parse_string(none)
            )
            self.assertFalse(Recovery.is_strict(none)[0], Schedule.parse_string(none))

            recovery = generate(3, ["a", "b", "c"], recovery="r")[0]
            self.assertTrue(
                Recovery.is_recoverable(recovery)[0], Schedule.parse_string(recovery)
            )
            self.assertFalse(
                Recovery.avoids_cascading_aborts(recovery)[0],
                Schedule.parse_string(recovery),
            )
            self.assertFalse(
                Recovery.is_strict(recovery)[0], Schedule.parse_string(recovery)
            )

            aca = generate(3, ["a", "b", "c"], recovery="a")[0]
            self.assertTrue(Recovery.is_recoverable(aca)[0], Schedule.parse_string(aca))
            self.assertTrue(
                Recovery.avoids_cascading_aborts(aca)[0], Schedule.parse_string(aca)
            )
            self.assertFalse(Recovery.is_strict(aca)[0], Schedule.parse_string(aca))

            st = generate(3, ["a", "b", "c"], recovery="s")[0]
            self.assertTrue(Recovery.is_recoverable(st)[0], Schedule.parse_string(st))
            self.assertTrue(
                Recovery.avoids_cascading_aborts(st)[0], Schedule.parse_string(st)
            )
            self.assertTrue(Recovery.is_strict(st)[0], Schedule.parse_string(st))

    def test_generate_deadlock(self):
        no_deadlock = 0
        for i in range(1000):
            deadlock = generate(3, ["a", "b", "c"], True)[0]
            self.assertTrue(predict_deadlock(deadlock), Schedule.parse_string(deadlock))
            y = generate(3, ["a", "b", "c"], False)[0]
            if predict_deadlock(y):
                no_deadlock += 1  # cases in which it not worked

    def test_scheduling_performer(self):
        test_caces_scheduling = [
            "w_1(z) w_3(y) r_1(x) r_3(z) r_2(y) w_2(z) w_3(x) c_1 c_2 c_3",
            "r_1(x) w_3(z) r_2(x) r_1(z) r_3(y) w_1(z) w_2(y) c_1 c_2 c_3",
            "r_1(x) w_3(z) r_2(x) r_1(z) r_3(y) w_1(z) w_2(y) r1(x) r1(x) c_1 c_2 c_3",
            "w2(x) r3(x) r1(y) r1(z) w3(y) w1(x) c3 w1(y) c1 r2(x) c2 ",
        ]
        for i in test_caces_scheduling:
            x = Schedule.parse_schedule(i)[0]
            self.assertTrue(
                Scheduling.is_S2PL(Perform_scheduling.perform_S2PL(x)[0]), "S2PL" + i
            )
            self.assertTrue(
                Scheduling.is_SS2PL(Perform_scheduling.perform_SS2PL(x)[0]), "SS2PL" + i
            )
            self.assertTrue(
                Scheduling.is_C2PL(Perform_scheduling.perform_C2PL(x)), "C2PL" + i
            )

    def test_compute_conflictset(self):
        # A tuple denotes: (schedule, conflictset)
        conflict_schedules = [
            (
                "w1(x)r2(y)r1(x)r2(x)c1w2(x)c2",
                [("w1(x)", "r2(x)"), ("w1(x)", "w2(x)"), ("r1(x)", "w2(x)")],
            ),
            ("w1(x)r2(y)r1(x)r2(x)c1w2(x)a2", []),
            ("w1(x)r2(y)r1(x)r2(x)a1w2(x)c2", []),
            ("w1(x)r2(y)r1(x)r2(x)a1w2(x)a2", []),
        ]
        for schedule, cset in conflict_schedules:
            solution = Perform_conflictgraph.compute_conflict_quantity(
                Schedule.parse_schedule(schedule)[0]
            )
            self.assertEqual(cset, solution, schedule)



from dbis_tm.TM import (
    Schedule,
    Operation,
    OperationType,
    ConflictGraph,
    ConflictGraphNode,
)
from typing import Union
from graphviz import Digraph
import copy


class Perform_scheduling:
    """
    I am a class which performs the schedulers.

    Functions:
        perform_C2PL (performs conservative 2-phase-locking on schedule)
        perform_S2PL (performs strict 2-phase-locking on schedule)
        perform_SS2PL (performs strong strict 2-phase-locking on schedule)
    """

    @classmethod
    def perform_S2PL(cls, schedule: Union[Schedule, str]) -> tuple[Schedule, str]:
        """
        This method performs S2PL.
        - No locks allowed after first unlock.
        - All unlocks have to be performed immediately after last operation.
        - Hold all writelocks till last r/w operation.
        Takes:  schedule
        Returns:    the locked schedule
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]
        schedule = copy.deepcopy(schedule)
        operations = []
        aborts = dict()
        commits = dict()
        transl = list([0] for i in range(schedule.tx_count))
        try_next_time = []
        counter = 1
        z = len(schedule.operations) * len(schedule.operations)
        el = 0
        len_operations = len(schedule.operations)
        # create a list for all actions of transactions, without duplicates
        for i in range(
            schedule.tx_count
        ):  # fix if the transactions are not correct numbered
            transl[i] = len(
                list(filter(lambda op: op.tx_number == (i + 1), schedule.operations))
            )
        last = []

        for x in range(schedule.tx_count):
            help1 = []
            [
                help1.append(op)
                for op in schedule.operations
                if op.tx_number == x + 1 and op not in help1
            ]
            last.append(help1[-1])  # problem unlocks of double locks not allowed
        while z > 0:
            j = schedule.operations[el]
            cur_op = j.op_type.value  # value1, _,
            cur_res = j.resource  # value2
            cur_trans = j.tx_number  # key
            # if conflict:
            op_waiting = list(
                filter(lambda op: op.tx_number == cur_trans, try_next_time)
            )
            concl = commits.keys()
            values_read = list(
                filter(
                    lambda op: op.op_type.value == "rl"
                    and op.tx_number != cur_trans
                    and op.tx_number not in concl
                    and cur_res == op.resource,
                    operations,
                )
            )
            values_write = list(
                filter(
                    lambda op: op.op_type.value == "wl"
                    and op.resource == cur_res
                    and op.tx_number not in concl
                    and op.tx_number != cur_trans,
                    operations,
                )
            )

            if (
                values_read and "r" != cur_op or values_write or op_waiting
            ):  # if already locked and current "r" operation, if other transactions, if already locked if other transactions, if in try next time
                try_next_time.append(j)
            else:
                locked = list(
                    filter(
                        lambda op: op.tx_number == cur_trans
                        and op.op_type.value == cur_op + "l"
                        and cur_res == op.resource,
                        operations,
                    )
                )
                if not locked:
                    operations.append(
                        Operation(
                            OperationType(cur_op + "l"), cur_trans, cur_res, counter
                        )
                    )
                    counter += 1
                operations.append(
                    Operation(OperationType(cur_op), cur_trans, cur_res, counter)
                )
                counter += 1
                transl[cur_trans - 1] -= 1
                if (
                    "r" == cur_op
                    and j.index
                    >= list(filter(lambda op: op.tx_number == cur_trans, last))[0].index
                    and not list(
                        filter(
                            lambda op: op == j,
                            schedule.operations[schedule.operations.index(j) + 1 :],
                        )
                    )
                ):
                    operations.append(
                        Operation(
                            OperationType(cur_op + "u"), cur_trans, cur_res, counter
                        )
                    )
                    counter += 1
                    # check for new actions possible
                    schedule.operations = try_next_time + schedule.operations[el + 1 :]
                    try_next_time = []
                    el = -1

            if transl[cur_trans - 1] == 0:  # check if unlock possible
                # problem perform unlocks of only the write ones
                unlock = list(
                    filter(
                        lambda op: op.tx_number == cur_trans
                        and (
                            "wl" == op.op_type.value
                            or (
                                "rl" == op.op_type.value
                                and Operation(
                                    OperationType("ru"), cur_trans, op.resource, counter
                                )
                                not in operations
                            )
                        ),
                        operations,
                    )
                )
                # filter all locks of cur_trans, only values
                for k in unlock:
                    operations.append(
                        Operation(
                            OperationType(k.op_type.value.replace("l", "u")),
                            cur_trans,
                            k.resource,
                            counter,
                        )
                    )
                    counter += 1
                if transl[cur_trans - 1] == 0:
                    commits[cur_trans] = counter
                    counter += 1
                # check for new actions possible
                schedule.operations = try_next_time + schedule.operations[el + 1 :]
                try_next_time = []
                el = -1
            z -= 1
            el += 1
            if el >= len(schedule.operations):
                schedule.operations = try_next_time
                try_next_time = []
                el = 0
            if len(schedule.operations) == 0:
                break
        all_op = list(filter(lambda op: op.op_type.value in ["r", "w"], operations))
        if (
            len(all_op) + len(commits) + len(aborts)
            != len_operations + schedule.tx_count
        ):
            return (
                Schedule(
                    operations, schedule.resources, schedule.tx_count, aborts, commits
                ),
                "Deadlock occured",
            )
        return (
            Schedule(
                operations, schedule.resources, schedule.tx_count, aborts, commits
            ),
            "",
        )

    @classmethod
    def perform_SS2PL(cls, schedule: Union[Schedule, str]) -> tuple[Schedule, str]:
        """
        Method to perform an SS2PL lock.
        - No locks allowed after first unlock.
        - All unlocks have to be performed immediately after last operation.
        - Hold all locks till last r/w operation.
        Takes:  schedule
        Returns:    the locked schedule

        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]
        schedule = copy.deepcopy(schedule)
        len_operations = len(schedule.operations)
        operations = []
        aborts = dict()
        commits = dict()
        transl = list([0] for i in range(schedule.tx_count))
        try_next_time = []
        counter = 1
        z = len(schedule.operations) * len(schedule.operations)
        el = 0

        for i in range(
            schedule.tx_count
        ):  # fix if the transactions are not correct numbered
            transl[i] = len(
                list(filter(lambda op: op.tx_number == (i + 1), schedule.operations))
            )

        while z > 0:
            j = schedule.operations[el]
            cur_op = j.op_type.value  # value1, _,
            cur_res = j.resource  # value2
            cur_trans = j.tx_number  # key
            # if conflict:
            op_waiting = list(
                filter(lambda op: op.tx_number == cur_trans, try_next_time)
            )
            concl = commits.keys()
            values_read = list(
                filter(
                    lambda op: op.op_type.value == "rl"
                    and op.tx_number != cur_trans
                    and op.tx_number not in concl
                    and cur_res == op.resource,
                    operations,
                )
            )
            values_write = list(
                filter(
                    lambda op: op.op_type.value == "wl"
                    and op.resource == cur_res
                    and op.tx_number not in concl
                    and op.tx_number != cur_trans,
                    operations,
                )
            )

            if (
                (values_read and "r" != cur_op) or values_write or op_waiting
            ):  # if already locked and current "r" operation, if other transactions, if already locked if other transactions, if in try next time
                try_next_time.append(j)
            else:
                locked = list(
                    filter(
                        lambda op: op.tx_number == cur_trans
                        and op.op_type.value == cur_op + "l"
                        and cur_res == op.resource,
                        operations,
                    )
                )
                if not locked:
                    operations.append(
                        Operation(
                            OperationType(cur_op + "l"), cur_trans, cur_res, counter
                        )
                    )
                    counter += 1
                operations.append(
                    Operation(OperationType(cur_op), cur_trans, cur_res, counter)
                )
                counter += 1
                transl[cur_trans - 1] -= 1

            if transl[cur_trans - 1] == 0:  # check if unlock possible
                # perform unlocks
                unlock = list(
                    filter(
                        lambda op: op.tx_number == cur_trans
                        and "l" in op.op_type.value,
                        operations,
                    )
                )  # filter all locks of cur_trans, only values
                for k in unlock:
                    operations.append(
                        Operation(
                            OperationType(k.op_type.value.replace("l", "u")),
                            cur_trans,
                            k.resource,
                            counter,
                        )
                    )
                    counter += 1
                commits[cur_trans] = counter
                schedule.operations = try_next_time + schedule.operations[el + 1 :]
                try_next_time = []
                el = -1
                counter += 1
            z -= 1
            el += 1
            if el >= len(schedule.operations):
                schedule.operations = try_next_time
                try_next_time = []
                el = 0
            if len(schedule.operations) == 0:
                break

        all_op = list(filter(lambda op: op.op_type.value in ["r", "w"], operations))
        if (
            len(all_op) + len(commits) + len(aborts)
            != len_operations + schedule.tx_count
        ):
            return (
                Schedule(
                    operations, schedule.resources, schedule.tx_count, aborts, commits
                ),
                "Deadlock occured",
            )
        return (
            Schedule(
                operations, schedule.resources, schedule.tx_count, aborts, commits
            ),
            "",
        )

    @classmethod
    def perform_C2PL(cls, schedule: Union[Schedule, str]) -> Schedule:
        """
        This method performs C2PL.
        - No locks allowed after first unlock.
        - All unlocks have to be performed immediately after last operation.
        - Hold all locks before performing an operation.
        Takes:  schedule
        Returns:    the locked schedule
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        schedule = copy.deepcopy(schedule)
        operations = []
        aborts = dict()
        commits = dict()
        transl = list([0] for i in range(schedule.tx_count))
        first = []
        try_next_time = []
        counter = 1
        z = len(schedule.operations) * len(schedule.operations)
        el = 0
        # create a list for all actions of transactions, without duplicates
        for i in range(
            schedule.tx_count
        ):  # fix if the transactions are not correct numbered
            transl[i] = len(
                list(filter(lambda op: op.tx_number == (i + 1), schedule.operations))
            )

        last = []
        for x in range(schedule.tx_count):
            help1 = []
            [
                help1.append(op)
                for op in schedule.operations
                if op.tx_number == x + 1 and op not in help1
            ]
            last.append(
                help1[-1]
            )  # to solve the problem of unlocks of double locks not allowed
        while z > 0:
            j = schedule.operations[el]
            cur_op = j.op_type.value  # value1, _,
            cur_res = j.resource  # value2
            cur_trans = j.tx_number  # key
            # if conflict:
            op_waiting = list(
                filter(lambda op: op.tx_number == cur_trans, try_next_time)
            )
            concl = commits.keys()

            read_lock = list(
                set(
                    [
                        op.resource
                        for op in operations
                        if op.tx_number != cur_trans
                        and op.tx_number not in concl
                        and op.op_type.value == "rl"
                    ]
                )
            )
            write_lock = list(
                set(
                    [
                        op.resource
                        for op in operations
                        if op.op_type.value == "wl"
                        and op.tx_number != cur_trans
                        and op.tx_number not in concl
                    ]
                )
            )

            first = list(
                filter(
                    lambda op: op.tx_number == cur_trans and op.index < j.index,
                    schedule.operations,
                )
            )

            if op_waiting:
                try_next_time.append(j)
            elif not first:  # first action of this trans
                all_op = list(
                    filter(lambda op: op.tx_number == cur_trans, schedule.operations)
                )
                write_need = list(
                    set([op.resource for op in all_op if op.op_type.value == "w"])
                )
                read_need = list(
                    set([op.resource for op in all_op if op.op_type.value == "r"])
                )
                locks = list(set(read_lock + write_lock))
                if [res for res in write_need if res in locks] or [
                    res for res in read_need if res in write_lock
                ]:  # w mit r&w checken, r mit w checken
                    try_next_time.append(j)
                else:  # perform locks
                    for l in all_op:
                        locked = list(
                            filter(
                                lambda op: op.tx_number == l.tx_number
                                and op.op_type.value == l.op_type.value + "l"
                                and l.resource == op.resource,
                                operations,
                            )
                        )
                        if not locked:
                            operations.append(
                                Operation(
                                    OperationType(l.op_type.value + "l"),
                                    l.tx_number,
                                    l.resource,
                                    counter,
                                )
                            )
                            counter += 1
                    operations.append(
                        Operation(OperationType(cur_op), cur_trans, cur_res, counter)
                    )
                    counter += 1
                    transl[cur_trans - 1] -= 1

            else:
                operations.append(
                    Operation(OperationType(cur_op), cur_trans, cur_res, counter)
                )
                counter += 1
                transl[cur_trans - 1] -= 1
                if (
                    "r" == cur_op
                    and j.index
                    >= list(filter(lambda op: op.tx_number == cur_trans, last))[0].index
                    and not list(
                        filter(
                            lambda op: op == j,
                            schedule.operations[schedule.operations.index(j) + 1 :],
                        )
                    )
                ):
                    operations.append(
                        Operation(
                            OperationType(cur_op + "u"), cur_trans, cur_res, counter
                        )
                    )
                    counter += 1
                    # check for new actions possible
                    schedule.operations = try_next_time + schedule.operations[el + 1 :]
                    try_next_time = []
                    el = -1

            if transl[cur_trans - 1] == 0:  # check if unlock possible
                unlock = list(
                    filter(
                        lambda op: op.tx_number == cur_trans
                        and (
                            "wl" == op.op_type.value
                            or (
                                "rl" == op.op_type.value
                                and Operation(
                                    OperationType("ru"), cur_trans, op.resource, counter
                                )
                                not in operations
                            )
                        ),
                        operations,
                    )
                )
                # filter all locks of cur_trans, only values
                for k in unlock:
                    operations.append(
                        Operation(
                            OperationType(k.op_type.value.replace("l", "u")),
                            cur_trans,
                            k.resource,
                            counter,
                        )
                    )
                    counter += 1
                if transl[cur_trans - 1] == 0:
                    commits[cur_trans] = counter
                    counter += 1
                # check for new actions possible
                schedule.operations = try_next_time + schedule.operations[el + 1 :]
                try_next_time = []
                el = -1
            z -= 1
            el += 1
            if el >= len(schedule.operations):
                schedule.operations = try_next_time
                try_next_time = []
                el = 0
            if len(schedule.operations) == 0:
                break

        return Schedule(
            operations, schedule.resources, schedule.tx_count, aborts, commits
        )


def predict_deadlock(schedule: Union[Schedule, str]) -> bool:
    # can only perform 2SSPL if no deadlock occurs
    x, is_deadlock = Perform_scheduling.perform_SS2PL(schedule)
    if is_deadlock != "":
        return True
    return False


class Perform_conflictgraph:
    @classmethod
    def compute_conflict_quantity(
        cls, schedule: Union[Schedule, str]
    ) -> tuple[str, list]:
        """
        Method to perform the omputing of the conflict set.
        No check for the corectnes of the schedule included. Please check this before using.
        Takes:
        - A schedule
        Reutrns:
        - The conflict set
        """
        if isinstance(schedule, str):
            schedule = Schedule.parse_schedule(schedule)
            assert not schedule[1]
            schedule = schedule[0]

        conflict_list_dup = []
        op = schedule.operations
        check = False
        aborted = list(schedule.aborts.keys())

        for i in op:  # generate conflict list
            check = False
            cur_trans = i.tx_number
            cur_res = i.resource
            if cur_trans in aborted:  # no aborted transactions allowed
                continue
            for j in op:
                if i == j:  # only start to check after the current operation
                    check = True
                    continue
                if check:
                    if (
                        j.tx_number not in aborted
                        and j.tx_number != cur_trans
                        and j.resource == cur_res
                    ):
                        # transaction not aborted, only conflict if not the same tansaction, but same resource
                        if (
                            i.op_type == OperationType.READ and i.op_type == j.op_type
                        ):  # no conflict if both are read operations
                            continue
                        conflict_list_dup.append([i, j])
        # discard duplicates
        conflict_list = []
        [conflict_list.append(x) for x in conflict_list_dup if x not in conflict_list]
        # compute string
        conflict = []
        for [k, l] in conflict_list:
            str1 = k.op_type.value + str(k.tx_number) + "(" + k.resource + ")"
            str2 = l.op_type.value + str(l.tx_number) + "(" + l.resource + ")"
            conflict_tuple = (str1, str2)
            conflict.append(conflict_tuple)
            # [("w_3(y)", "w_1(y)"), ("w_1(y)", "r_3(y)"), ("w_1(z)", "w_3(z)"), ("w_1(x)", "r_3(x)")]
        return conflict

    @classmethod
    def compute_conflictgraph(cls, conflict_list: dict) -> ConflictGraph:
        knots = []
        for i in range(1, len(conflict_list) + 1):
            knots.append(ConflictGraphNode(i))
        graph = ConflictGraph()
        print("test:", conflict_list)
        for i in conflict_list.keys():
            if conflict_list[i] != set():
                for k in conflict_list[i]:
                    graph.add_edge(knots[i - 1], knots[k - 1])
        return graph



"""
Created 2022-05

@author: Lara
@author: Marc
@author: Wolfgang
"""
from dbis_tm.TM import Schedule, ConflictGraph, ConflictGraphNode
from dbis_tm.TMSolver import Recovery, Scheduling, Serializability
from dbis_tm.TMCheck import SyntaxCheck, ScheduleCheck
from tests.scheduletest import ScheduleTest


class TestTM(ScheduleTest):
    """
    tests and playground
    for https://git.rwth-aachen.de/i5/teaching/dbis-digi-2022/-/issues/35
    """

    # A tuple denotes: (schedule, is_well_formed, is_serializable)
    unparsed_schedule_tests = [
        ("w1(x)w1(y)r2(u)w2(x)r2(y)r3(x)w2(z)a2r1(z)c1c3", True, True),
        # well-formed, serializable
        ("R1(x)W1(x)r2(x)A1w2(x)C2", True, True),
        # well-formed, serializable
        ("w1(Y)w4(W)w3(Z)w2(X)r1(Z)r1(X)r4(Z)", True, True),
        # well-formed, serializable
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1(y)w2(z)w1(z)w3(y)r2(x)c3r2(y)c2w1(y)a1",
            True,
            False,
        ),
        # well-formed, not serializable
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1(y)w2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            True,
            False,
        ),
        # well-formed, not serializable
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1y)w2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            False,
            False,
        ),
        # malformed
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w(y)w2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            False,
            False,
        ),
        # malformed
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1(yw2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            False,
            False,
        ),
        # malformed
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)1(y)w2(z)w1(z)w3(y)c3r2(y)c2w1(y)c1",
            False,
            False,
        )
        # malformed
    ]

    # A tuple denotes: (schedule, is_RC, is_ACA, is_ST)
    schedule_recoverability_tests = [
        ("w1(x)r2(y)r1(x)r2(x)c1w2(x)c2", True, False, False),
        ("w1(x)w2(y)r1(y)r2(x)c2w1(y)c1", False, False, False),
        ("w1(x)r2(y)c1w2(x)r2(x)w2(y)c2", True, True, True),
        ("w2(x)r1(y)r2(x)r1(x)c2w1(y)c1", True, False, False),
        ("w1(x)w1(y)r2(u)w2(x)r2(y)w2(y)c2w1(z)c1", False, False, False),
        ("w1(x)w1(y)r2(u)w2(x)r2(y)w2(y)w1(z)c1c2", True, False, False),
        ("w1(x)w1(y)r2(u)w2(x)r2(y)r3(x)w2(z)a2r1(z)c1c3", False, False, False),
        (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1(y)w2(z)w1(z)w3(y)r2(x)c3r2(y)c2w1(y)a1",
            True,
            False,
            False,
        ),
        ("w1(x)r2(y)r1(x)r2(x)c1w2(x)c2", True, False, False),
        ("w1(x)w2(y)r1(y)r2(x)c2w1(y)c1", False, False, False),
        ("w1(x)r2(y)c1w2(x)r2(x)w2(y)c2", True, True, True),
        ("w2(x)r1(y)r2(x)r1(x)c2w1(y)c1", True, False, False),
        ("w1(x)w1(y)r2(u)w2(x)r2(y)w2(y)w1(z)c1c2", True, False, False),
        ("w1(x)w1(y)r2(u)w2(x)w1(z)c1r2(y)w2(y)c2", True, True, False),
        ("w1(x)w1(y)r2(u)w1(z)c1w2(x)r2(y)w2(y)c2", True, True, True),
    ]

    # A tuple denotes: (schedule, is_2PL, is_C2PL, is_S2PL, is_SS2PL)
    scheduling_tests = [
        (
            "rl2(e)wl1(x)w1(x)r2(e)rl1(y)r1(y)ru1(y)wu1(x)ru2(e)",
            True,
            False,
            True,
            True,
        ),
        ("wl1(x)w1(x)r2(e)wu1(x)rl1(y)r1(y)ru1(y)", False, False, False, False),
        ("wl1(x)rl1(y)w1(x)r1(y)ru1(y)wu1(x)", True, True, True, True),
        ("wl1(x)w1(x)rl1(y)r1(y)ru1(y)wu1(x)", True, False, True, True),
        ("wl1(x)rl1(y)w1(x)wu1(x)r1(y)ru1(y)", True, True, False, False),
        ("wl1(x)rl1(y)w1(x)r1(y)ru1(y)wu1(x)", True, True, True, True),
        ("wl1(x)rl1(y)r1(y)ru1(y)w1(x)wu1(x)", True, True, True, False),
        ("rl2(y)r2(y)wl3(x)w3(x)wl1(z)w1(z)", True, True, True, True),
        (
            "rl1(x)r1(x)wl3(z)w3(z)rl2(x)r2(x)rl3(y)r3(y)wu3(z)ru3(y)"
            "c3rl1(z)r1(z)wl1(z)ru1(x)ru1(z)w1(z)wu1(z)c1wl2(y)ru2(x)"
            "w2(y)wu2(y)c2",
            True,
            False,
            True,
            False,
        ),
        (
            "wl1(z)rl1(x)w1(z)wu1(z)r1(x)ru1(x)c1wl3(y)rl3(z)wl3(x)"
            "w3(y)wu3(y)r3(z)ru3(z)rl2(y)wl2(z)r2(y)ru2(y)w2(z)wu2(z)"
            "c2w3(x)wu3(x)c3",
            True,
            True,
            False,
            False,
        ),
    ]

    # A tuple denotes: (schedule, schedule_mod, compare_schedules)
    compare_schedules_test = [
        (
            "w1(x)r2(e)r1(y)",
            "rl2(e)wl1(x)w1(x)r2(e)rl1(y)r1(y)ru1(y)wu1(x)ru2(e)",
            True,
        ),
        (
            "r2(y)w3(x)w1(z)w3(y)r1(x)r2(z)r3(z)c1c2c3",
            "rl2(y)r2(y)wl3(x)w3(x)wl1(z)w1(z)wl3(y)w3(y)rl1(x)r1(x)wu1(z)ru1(x)rl2(z)r2(z)ru2(z)ru2(y)"
            "rl3(z)r3(z)wu3(x)wu3(y)ru3(z)c1c2c3",
            True,
        ),
    ]

    def test_schedule_parsing(self):
        """
        tests parse_schedule(unparsed_schedule)
        """
        for (schedule, is_well_formed, _), i in zip(
            self.unparsed_schedule_tests, range(0, len(self.unparsed_schedule_tests))
        ):
            _, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(is_well_formed, msg == "", f"Schedule {i}:")

    def test_serializability(self):
        """
        tests is_serializable(parsed_schedule)
        """
        for (schedule, _, is_serializable), i in zip(
            self.unparsed_schedule_tests, range(len(self.unparsed_schedule_tests))
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            if msg == "":
                Serializability.remove_aborted_tx(parsed)
                actual, _ = Serializability.is_serializable(parsed)
                self.assertEqual(is_serializable, actual, f"Schedule {i}:")

    def test_removing_aborted_TM(self):
        """
        tests test_remove_aborted_TM(parsed_schedule)
        """
        unparsed_schedule = (
            "r1(x)w2(y)r1(x)w3(z)w3(x)r1(y)w1(y)w2(z)w1(z)w3(y)r2(x)c3r2(y)c2w1(y)a1"
        )
        expected = "w2(y)w3(z)w3(x)w2(z)w3(y)r2(x)r2(y)"

        parsed, msg = Schedule.parse_schedule(unparsed_schedule)
        self.assertEqual(msg, "")

        actual, _ = Serializability.is_serializable(parsed)
        self.assertFalse(actual)

        Serializability.remove_aborted_tx(parsed)
        expected, msg = Schedule.parse_schedule(expected)
        self.assertEqual(msg, "")
        actual, _ = Serializability.is_serializable(expected)
        self.assertFalse(actual)
        self.assertEqual(expected.operations, parsed.operations)

        actual, _ = Serializability.is_serializable(parsed)
        self.assertFalse(actual)

    def test_reads_from(self):
        """
        tests reads_from(parsed_schedule, tx1, resource, tx2)
        """
        unparsed_schedule = "w1(y)r2(y)a1c2"
        parsed1, msg = Schedule.parse_schedule(unparsed_schedule)
        self.assertEqual(msg, "")
        self.assertTrue(Recovery.reads_from(parsed1, 2, "y", 1)[0])

        unparsed_schedule = "w1(x)w1(y)r2(u)w2(x)r2(y)r3(x)w2(z)a2r1(z)c1c3"
        parsed2, msg = Schedule.parse_schedule(unparsed_schedule)
        self.assertEqual(msg, "")
        self.assertTrue(Recovery.reads_from(parsed2, 3, "x", 2)[0])
        self.assertFalse(Recovery.reads_from(parsed2, 3, "x", 1)[0])
        self.assertTrue(Recovery.reads_from(parsed2, 2, "y", 1)[0])
        self.assertFalse(Recovery.reads_from(parsed2, 1, "z", 2)[0])

        unparsed_schedule = "w1(y)a1r2(y)c2"
        parsed1, msg = Schedule.parse_schedule(unparsed_schedule)
        self.assertEqual(msg, "")
        self.assertFalse(Recovery.reads_from(parsed1, 2, "y", 1)[0])

        unparsed_schedule = "w3(y)w1(y)r2(y)c2a1c3"
        parsed1, msg = Schedule.parse_schedule(unparsed_schedule)
        self.assertEqual(msg, "")
        self.assertTrue(Recovery.reads_from(parsed1, 2, "y", 1)[0])
        self.assertFalse(Recovery.reads_from(parsed1, 2, "y", 3)[0])

        unparsed_schedule = "w3(y)w1(y)a1r2(y)c2c3"
        parsed1, msg = Schedule.parse_schedule(unparsed_schedule)
        self.assertEqual(msg, "")
        self.assertFalse(Recovery.reads_from(parsed1, 2, "y", 1)[0])
        self.assertTrue(Recovery.reads_from(parsed1, 2, "y", 3)[0])

    def test_is_recoverable(self):
        """
        tests is_recoverable(parsed_schedule)
        """
        for (schedule, is_recoverable, _, _), i in zip(
            self.schedule_recoverability_tests,
            range(len(self.schedule_recoverability_tests)),
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(msg, "")
            actual, result = Recovery.is_recoverable(parsed)
            # TODO: Check result (i.e., counterexample or proof)
            self.assertEqual(is_recoverable, actual, f"Schedule {i}: {schedule}")

    def test_avoids_cascading_aborts(self):
        """
        tests avoids_cascading_aborts(parsed_schedule)
        """
        for (schedule, _, avoids_cascading_aborts, _), i in zip(
            self.schedule_recoverability_tests,
            range(len(self.schedule_recoverability_tests)),
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(msg, "")
            actual, result = Recovery.avoids_cascading_aborts(parsed)
            # TODO: Check result (i.e., counterexample or proof)
            self.assertEqual(
                avoids_cascading_aborts, actual, f"Schedule {i}: {schedule}"
            )

    def test_is_strict(self):
        """
        tests is_strict(parsed_schedule)
        """
        for (schedule, _, _, is_strict), i in zip(
            self.schedule_recoverability_tests,
            range(len(self.schedule_recoverability_tests)),
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(msg, "")
            actual, result = Recovery.is_strict(parsed)
            # TODO: Check result (i.e., counterexample or proof)
            self.assertEqual(is_strict, actual, f"Schedule {i}: {schedule}")

    def test_is_2PL(self):
        """
        tests is_2PL(parsed_schedule)
        """
        for (schedule, is2PL, _, _, _), i in zip(
            self.scheduling_tests, range(len(self.scheduling_tests))
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(msg, "")
            actual, errors = Scheduling.is_2PL(parsed)
            self.assertEqual(
                is2PL, actual, f"Schedule {i}: {schedule}\n Error: {errors}"
            )
            self.assertEqual(not errors, is2PL)
            # TODO: Check for correct error messages in case of is2PL == False

    def test_is_C2PL(self):
        """
        tests is_C2PL(parsed_schedule)
        """
        for (schedule, _, isC2PL, _, _), i in zip(
            self.scheduling_tests, range(len(self.scheduling_tests))
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(msg, "")
            actual, errors = Scheduling.is_C2PL(parsed)
            self.assertEqual(
                isC2PL, actual, f"Schedule {i}: {schedule}\n Error: {errors}"
            )
            self.assertEqual(not errors, isC2PL)
            # TODO: Check for correct error messages in case of isC2PL == False

    def test_is_S2PL(self):
        """
        tests is_S2PL(parsed_schedule)
        """
        for (schedule, _, _, isS2PL, _), i in zip(
            self.scheduling_tests, range(len(self.scheduling_tests))
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(msg, "")
            actual, errors = Scheduling.is_S2PL(parsed)
            self.assertEqual(
                isS2PL, actual, f"Schedule {i}: {schedule}\n Error: {errors}"
            )
            # self.assertEqual(not errors, isS2PL)
            # TODO: Check for correct error messages in case of isS2PL == False

    def test_is_SS2PL(self):
        """
        tests is_SS2PL(parsed_schedule)
        """
        for (schedule, _, _, _, isSS2PL), i in zip(
            self.scheduling_tests, range(len(self.scheduling_tests))
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            self.assertEqual(msg, "")
            actual, errors = Scheduling.is_SS2PL(parsed)
            self.assertEqual(
                isSS2PL, actual, f"Schedule {i}: {schedule}\n Error: {errors}"
            )
            # self.assertEqual(not errors, isSS2PL)
            # TODO: Check for correct error messages in case of isSS2PL == False

    def test_compare_schedules(self):
        """
        tests check_schedule(schedule, schedule_mod)
        """
        for (schedule, schedule_mod, result), _i in zip(
            self.compare_schedules_test, range(len(self.compare_schedules_test))
        ):
            parsed, msg = Schedule.parse_schedule(schedule)
            parsed_mod, msg_mod = Schedule.parse_schedule(schedule_mod)
            returned = Schedule.is_operations_same(parsed, parsed_mod)
            # returned =len(problems)==0
            self.assertEqual(returned, result)

    def testEdgeLessConflictGraph(self):
        """
        test the content of an empty graph
        """
        g_1 = ConflictGraph()
        t1 = ConflictGraphNode(1)
        t2 = ConflictGraphNode(2)
        self.assertTrue(g_1.isEmpty())
        gvMarkup = g_1.get_graphviz_graph()
        debug = False
        if debug:
            print(gvMarkup)
        self.assertTrue(
            """{
	graph [label="Conflict Graph"]
}"""
            in str(gvMarkup)
        )
        g_1.add_edge(t1, t2)
        gvMarkup = g_1.get_graphviz_graph()
        if debug:
            print(gvMarkup)
        self.assertTrue("t1 -> t2" in str(gvMarkup))

    def testConfSyntaxCheck(self):
        """
        test the SyntaxCheck functionality for Conflicts
        """
        s1_conf = [("w_2(x)", "r_1(x)"), ("w_1(z)", "w_2(z)")]
        s2_conf = [
            ("r_2(x)", "w_3(x)"),
            ("w_1(y)", "r_2(y)"),
            ("w_1(y)", "w_2(y)"),
            ("w_1(y)", "w_3(y)"),
            ("w_3(z)", "w_1(z)"),
            ("w_3(z)", "w_2(z)"),
            ("r_2(y)", "w_3(y)"),
            ("w_2(y)", "w_3(y)"),
            ("w_2(y)", "r_1(y)"),
            ("w_1(z)", "w_2(z)"),
            ("w_3(y)", "r_1(y)"),
            ("w_3(y)", "w_2(y)"),
            ("r_1(y)", "w_2(y)"),
        ]
        conf_err1 = {}
        conf_err2 = []
        conf_err3 = "Garbage"
        conf_err4 = [("a"), ("b", "c", "e")]
        conf_err5 = [("a_3(x)", "b")]
        debug = False
        expectedList = [
            None,
            None,
            "{} ist keine Liste",
            None,
            "Garbage ist keine Liste",
            (
                "Das Tupel ('b', 'c', 'e') von [('b', 'c', 'e'), 'a'] ist kein Paar",
                "Das Tupel a von ['a', ('b', 'c', 'e')] ist kein Paar",
            ),
            "Das Tupel ('a_3(x)', 'b') von [('a_3(x)', 'b')] hat keine korrekte Syntax",
        ]
        for i, conf in enumerate(
            [s1_conf, s2_conf, conf_err1, conf_err2, conf_err3, conf_err4, conf_err5]
        ):
            msg = SyntaxCheck.check_conf_set_syntax(conf)
            if debug:
                print(f"{i}:{msg}", "test")
            expected = expectedList[i]
            if isinstance(expected, tuple):
                expected1, expected2 = expected
                self.assertTrue(expected1 == msg or expected2 == msg)
            else:
                self.assertEqual(expected, msg)

    def testScheduleSyntaxCheck(self):
        """
        test the SyntaxCheck functionality for Schedules
        """
        schedule_1 = "w_2(x) w_1(z) r_2(y) r_1(x) r_3(z) w_3(x) w_1(y) c_1 c_2 c_3"
        schedule_2 = "r_2(z) w_1(y) r_3(z) r_2(y) r_1(x) w_2(y) w_3(x) c_1 c_2 c_3"
        schedule_3 = "r_1(x) w_2(z) w_3(y) w_2(x) r_3(z) r_1(y) r_2(y) c_1 c_2 c_3"
        schedule_4 = "w2(x)"
        schedule_5 = "rl_1(x) r_1(x) wl_2(z) w_2(z)  wl_3(y) w_3(y) wl_2(x) w_2(x) rl_3(z) r_3(z) wu_3(y) ru_3(z) rl_1(y) r_1(y) ru_1(x) ru_1(y) rl_2(y) r_2(y) wu_2(z) wu_2(x) ru_2(y) c_1 c_2 c_3"
        schedule_6 = "wl_2(x) rl_2(y) w_2(x) wu_2(x) r_2(y) ru_2(y) c_2 wl_1(z) rl_1(x) wl_1(y) w_1(z) wu_1(z) r_1(x) ru_1(x) rl_3(z) wl_3(x) r_3(z) ru_3(z) w_3(x) wu_3(x) c_3 w_1(y) wu_1(y) c_1"
        schedule_err2 = ""
        expected = [
            None,
            None,
            None,
            None,
            None,
            None,
            "Leerer Schedule kann keine Lösung sein",
        ]
        debug = False
        for i, schedule in enumerate(
            [
                schedule_1,
                schedule_2,
                schedule_3,
                schedule_4,
                schedule_5,
                schedule_6,
                schedule_err2,
            ]
        ):
            msg = SyntaxCheck.check_schedule_syntax(schedule)
            if debug:
                print(msg)
            self.assertEqual(expected[i], msg)



"""
Created on 2022-07-06

@author: Lara
@author: Marc
@author: wf

"""
import re
from dbis_tm.TM import Schedule
from dbis_tm.TMSolver import Scheduling, Recovery, Serializability
from dbis_tm.Solution_generator import Perform_conflictgraph
from typing import Union
from excmanager.scorer import Scorer, SetScorer


class ScheduleCheck:
    """
    check the schedule
    """

    @classmethod
    def check(cls, index, schedule, result) -> str:
        """
        check the given schedule against the given result
        """
        msg = None
        s_parsed, s_problem = Schedule.parse_schedule(schedule)
        result_parsed, result_problem = Schedule.parse_schedule(result)
        problems = Scheduling.check_operations_same(s_parsed, result_parsed)
        if not len(problems) == 0:
            msg = f"schedule_{index} enthält unterschiedliche oder nicht alle Operationen aus s{index}"
        return msg

    @classmethod
    def feedback(cls, msg: str, positive: bool):
        """
        give positive or negative feedback with the given message

        Args:
            msg(str): the message to print
            positive(bool): if True add a ✅ marker else ❌
        """
        marker = "✅" if positive else "❌"
        print(f"{msg} {marker}")

    @classmethod
    def check_same(
        self, original: Schedule, schedule: Schedule, schedule_str: str
    ) -> bool:
        """
        Checks whether the _schedule_ has the same operations (without locks) as _original_
            and whether _schedule_str_ can be parsed without errors

        Returns:
            True if no errors occur
        """
        sameProblemCount = 0
        problems = Schedule.is_operations_same(original, schedule)
        if not problems:
            self.feedback(
                f"schedule has not the same operations as the original.", False
            )
            sameProblemCount += 1
        else:
            self.feedback("schedule covers all operations", True)

        s_parsed, error_parse = Schedule.parse_schedule(schedule_str)
        if error_parse:
            ScheduleCheck.feedback(
                f"Could not parse schedule due to '{error_parse}'", False
            )
            sameProblemCount += 1
        else:
            ScheduleCheck.feedback("parse schedule ok", True)
        return sameProblemCount == 0


class SyntaxCheck:
    """
    I am an interface for checking the syntax of inputs.
    You should not construct me because I am a stateless interface that merely provides static functions.

    Functions:
        check_conf_set_syntax (checks syntax of strings in tuple that denotes conflicting operations)
    """

    def __init__(self):
        raise TypeError("Cannot create 'SyntaxCheck' instances.")

    @classmethod
    def check_schedule_syntax(cls, schedule: str) -> str:
        """
        check the syntax of the given schedule

        Args:
            schedule(str): the schedule to check

        Returns:
            msg: None if ok else the problem message
        """
        schedule = Schedule.sanitize(schedule)
        syntax_pattern = "([rw][lu]?[1-3][(][a-z][)]|[c][1-3])?"
        p_count = re.findall(syntax_pattern, schedule).count("")
        msg = None
        if schedule == "":
            msg = "Leerer Schedule kann keine Lösung sein"
        if p_count > 1:
            msg = f"Schedule '{schedule}' hat keine korrekte Syntax"
        return msg

    @classmethod
    def check_conf_set_syntax(cls, conf_set: list[tuple[str, str]]) -> str:
        """
        Check syntax of strings in tuple that denotes conflicting operations.

        Returns:
            None if input is formatted according to pattern
            or an error message in case a tuple is formatted incorrectly
        """
        tuple_pattern = "[rw][1-3][(][a-z][)]|[rw]_[1-3][(][a-z][)]"
        if not isinstance(conf_set, list):
            return f"{conf_set} ist keine Liste"
        for t in conf_set:
            if not len(t) == 2:
                return f"Das Tupel {t} von {conf_set} ist kein Paar"
            for s in sorted(list(t)):
                if not re.match(tuple_pattern, s):
                    return f"Das Tupel {t} von {conf_set} hat keine korrekte Syntax"
        return None


class ScheduleScorer(Scorer):
    """
    Class to score scheduling
    """

    def __init__(self, debug: bool = False):
        """
        Constructor

        Args:
            debug(bool): if True show debug information
        """
        Scorer.__init__(self, debug=debug)

    def check_schedule(self, parsed: Schedule, check: str) -> list[str]:
        """
        Checks whether the schedule given (parsed) is in the correct form(check)

        Returns:
         array of errors
            or empty array
        """
        if check == "C2PL":
            _presult, errors = Scheduling.is_C2PL(parsed)

        elif check == "S2PL":
            _presult, errors = Scheduling.is_S2PL(parsed)

        elif check == "SS2PL":
            _presult, errors = Scheduling.is_SS2PL(parsed)
        else:
            errors = []

        return errors

    def getScore(
        self, original: str, schedule: str, check_scheduling: str, max_points: int
    ) -> int:
        """
        - Run syntax check for schedule
        - Check schedule for _check_scheduling_
        - Add points

        Returns:
            points
        """
        original_p, original_error = Schedule.parse_schedule(original)
        schedule_p, schedule_error = Schedule.parse_schedule(schedule)
        checkSame = ScheduleCheck.check_same(original_p, schedule_p, schedule)
        if checkSame:
            check = f"checking '{check_scheduling}' of schedule '{schedule}'"
            errors = self.check_schedule(schedule_p, check_scheduling)
            if errors:
                for error in errors:
                    # just for feedback
                    self.addScore(max_points, check, problem=f"'{error}'")
            else:
                self.addScore(max_points, check, problem=None)
        return self.score


class ConflictSetScorer:
    """
    scorer for conflict sets
    """

    def removeBlanks(self, result):
        """
        remove blanks and underscores
        """
        resultNoBlanks = {
            (str1.replace(" ", ""), str2.replace(" ", "")) for (str1, str2) in result
        }
        resultNoBlanks = {
            (str1.replace("_", ""), str2.replace("_", "")) for (str1, str2) in result
        }
        return resultNoBlanks

    def score_conflictSet(self, result1, result2, schedule1, schedule2, max_points):
        """
        score the given result sets against the given solutions
        """
        result1 = self.removeBlanks(result1)
        solution1 = Perform_conflictgraph.compute_conflict_quantity(schedule1)
        setScorer1 = SetScorer()
        setScorer1.evaluate_set(result1, solution1, max_points=max_points / 2)
        result2 = self.removeBlanks(result2)
        solution2 = Perform_conflictgraph.compute_conflict_quantity(schedule2)
        setScorer2 = SetScorer()
        setScorer2.evaluate_set(result2, solution2, max_points=max_points / 2)
        score = setScorer1.score + setScorer2.score
        return round(score, 2)


class ConflictSerializationScorer(Scorer):
    """
    a scorer for conflict serialization tasks
    """

    def score_conflictSerialization(
        self, name, cgresult, serializableResult, schedule, max_points
    ):
        """
        score a conflict serialization answer
        """
        points_seri = 0.5
        points_graph = max_points - points_seri
        if points_graph <= 0:
            points_seri = max_points / 2
            points_graph = max_points / 2
        serializable = Serializability.is_serializable(schedule)
        cgsolution = Perform_conflictgraph.compute_conflictgraph(serializable[1])

        serializableCheck = f"check that your result for serializable of {name} {serializableResult} = correct solution serializable {serializable[0]}"
        serializableProblem = None
        if serializableResult != serializable[0]:
            serializableProblem = ""
        self.addScore(points_seri, serializableCheck, serializableProblem)
        expectedGraph = str(cgsolution.digraph)
        conflictGraphCheck = f"check that conflictGraph is '''{expectedGraph}'''"
        conflictGraphProblem = None
        if cgresult != cgsolution:
            conflictGraphProblem = f"{str(cgresult.digraph)} is different"
        self.addScore(points_graph, conflictGraphCheck, conflictGraphProblem)
        return self.score


class RecoveryScorer(Scorer):
    """
    a scorer for recovery
    """

    def removeBlanksAndUnderScores(self, proof):
        """
        sanitize the given proof set

        Args:
            proof(set): the set to sanitize
        Return:
            set: set with no _ or blank in notation
        """
        proofResult = set()
        # loop over the tuples
        for t in list(proof):
            tlist = list(t)
            rtuple = ()
            for telement in tlist:
                if isinstance(telement, str):
                    for charToRemove in [" ", "_"]:
                        telement = telement.replace(charToRemove, "")
                rtuple += (telement,)
            proofResult.add(rtuple)
        return proofResult

    def setString(self, aSet):
        if aSet == set():
            return "{}"
        else:
            return str(aSet)

    def score_proof(
        self, name, proofName, isClass, proofClass, isClassSolution, proofClassSolution
    ):
        """
        score a proof

        Args:
            name(str): the name of the schedule
            proofName(str): the name of the proof/Class e.g. RC/ACA/ST
            isClass(bool): result if the schedule is in the class
            proofClass(set): the set to prove with
            isClassSolution(bool): True if the schedule is in the given class as the sample solution
            proofClassSolution(set): the set which proves to be in the class
        """
        negativeProof = False
        problem = None
        proofClass = self.removeBlanksAndUnderScores(proofClass)
        # positive Proof
        if isClassSolution:
            check = f"{name} is {proofName} with expected proof {proofClassSolution}"
            # check sets:
            errors1 = []
            errors2 = []
            for i in proofClassSolution:
                if i not in proofClass:
                    errors1.append(i)
            for j in proofClass:
                if j not in proofClassSolution:
                    errors2.append(j)
            if isClassSolution == isClass and len(errors1) == 0 == len(errors2):
                pass
            else:
                problem = f"Missing tuples: {errors1},wrong tuples: {errors2}"
            self.addScore(self.points_per_class, check, problem)
        else:
            check = (
                f"{name} is not {proofName} with expected proof {proofClassSolution}"
            )
            if isClassSolution == isClass:
                # get first element of set
                if len(proofClassSolution) == 0 and len(proofClass) == 0:
                    negativeProof = True
                else:
                    if len(proofClass) > 0:
                        proofClassItem = tuple(proofClass)[0]
                        negativeProof = (
                            len(proofClass) > 0 and proofClassItem in proofClassSolution
                        )
            if not negativeProof:
                problem = (
                    f"your proof {self.setString(proofClass)} is not the expected one"
                )
            self.addScore(self.points_per_class, check, problem)
        return negativeProof

    def score_recovery(
        self, name: str, schedule: Union[Schedule, str], result, max_score
    ) -> int:
        """
        We grade each class.

        RC needs a proof. For ACA and ST we can use the subset relationship to avoid giving a counterexample.

        That is, we either give a counterexample
        or we simply leave the proof empty but set the bool to the correct value. In that case it is necessary to
        show the class we are relying on for the subset relationship. That is, if a schedule is not ST, we can use the
        subset relationship IF it was shown that the schedule is not ACA.
        """
        self.points_per_class = max_score * (1 / 3)
        is_in_class, proof = result
        is_rc, is_aca, is_st = is_in_class
        proof_rc, proof_aca, proof_st = proof
        # RC
        is_rc_solution, proof_rc_solution = Recovery.is_recoverable(schedule)
        showed_not_rc = self.score_proof(
            name, "RC", is_rc, proof_rc, is_rc_solution, proof_rc_solution
        )

        # ACA
        is_aca_solution, proof_aca_solution = Recovery.avoids_cascading_aborts(schedule)
        if showed_not_rc:
            proof_aca_solution = {}
        showed_not_aca = self.score_proof(
            name, "ACA", is_aca, proof_aca, is_aca_solution, proof_aca_solution
        )

        # ST
        is_st_solution, proof_st_solution = Recovery.is_strict(schedule)
        if showed_not_aca:
            proof_st_solution = {}
        self.score_proof(name, "ST", is_st, proof_st, is_st_solution, proof_st_solution)
        return self.score



from dbis_tm.Generate import generate
from dbis_tm.Solution_generator import (
    Perform_conflictgraph,
    Perform_scheduling,
    predict_deadlock,
)
from dbis_tm.TM import Schedule
from dbis_tm.TMSolver import Serializability, Recovery
import random
from IPython.display import display


class Generate_Tasks:
    def __init__(self):
        create = Creating()
        self.task1 = create.create_serialibility()
        self.task2 = create.create_recovery()
        self.task3 = create.create_scheduling()

    def get_tasks(self):
        serializability = [
            [self.task1.subtask1, self.task1.solution1],
            [self.task1.subtask2, self.task1.solution2],
        ]  # solution can not yet be generated
        recovery = [
            [self.task2.subtask1, self.task2.solution1],
            [self.task2.subtask2, self.task2.solution2],
            [self.task2.subtask3, self.task2.solution3],
        ]
        scheduling = [
            [self.task3.subtask1, self.task3.solution1],
            [self.task3.subtask2, self.task3.solution2],
            [self.task3.subtask3, self.task3.solution3],
        ]
        return serializability, recovery, scheduling


class Task:
    def __init__(
        self, task1, solution1, task2=None, solution2=None, task3=None, solution3=None
    ):
        self.subtask1 = task1
        self.solution1 = solution1
        self.subtask2 = task2
        self.solution2 = solution2
        self.subtask3 = task3
        self.solution3 = solution3


class Creating:
    def create_serialibility(self) -> Task:
        # check not empty
        for i in range(100):
            schedule1 = generate(3, ["a", "b", "c"])[0]
            schedule2 = generate(3, ["a", "b", "c"])[0]
            schedule1_solution = Perform_conflictgraph.compute_conflict_quantity(
                schedule1
            )
            schedule2_solution = Perform_conflictgraph.compute_conflict_quantity(
                schedule2
            )
            if len(schedule1_solution) >= 2 and len(schedule2_solution) >= 2:
                break

        for i in range(1000):
            schedule3 = generate(3, ["r", "u", "k"])[0]
            task3_seril = Serializability.is_serializable(schedule3)
            task3 = Perform_conflictgraph.compute_conflict_quantity(schedule3)
            if len(task3) > 5:
                break
        for i in range(1000):
            for i in range(1000):
                schedule4 = generate(3, ["r", "u", "k"])[0]
                if Serializability.is_serializable(schedule4) != task3_seril[0]:
                    break
            task4_seril = Serializability.is_serializable(schedule4)
            task4 = Perform_conflictgraph.compute_conflict_quantity(schedule4)
            if len(task4) > 5:
                break
        task3_solution = [
            task3_seril[0],
            Perform_conflictgraph.compute_conflictgraph(
                task3_seril[1]
            ).get_graphviz_graph,
        ]
        task4_solution = [
            task4_seril[0],
            Perform_conflictgraph.compute_conflictgraph(
                task4_seril[1]
            ).get_graphviz_graph,
        ]
        schedule1 = Schedule.parse_string(schedule1)[0]
        schedule2 = Schedule.parse_string(schedule2)[0]
        schedule3 = Schedule.parse_string(schedule3)[0]
        schedule4 = Schedule.parse_string(schedule4)[0]

        return Task(
            [schedule1, schedule2],
            [schedule1_solution, schedule2_solution],
            [schedule3, task3, schedule4, task4],
            [task3_solution, task4_solution],
        )

    def create_recovery(self) -> Task:
        options = ["n", "r", "a", "s"]
        perform = random.sample(options, 3)

        schedule1 = generate(3, ["a", "b", "c"], recovery=perform[0])[0]
        schedule2 = generate(3, ["a", "b", "c"], recovery=perform[1])[0]
        schedule3 = generate(3, ["a", "b", "c"], recovery=perform[2])[0]

        solution1_r = Recovery.is_recoverable(schedule1)
        if not solution1_r[0]:
            solution1_a = (False, {})
            solution1_s = (False, {})
        else:
            solution1_a = Recovery.avoids_cascading_aborts(schedule1)
        if not solution1_a[0]:
            solution1_s = (False, {})
        else:
            solution1_s = Recovery.is_strict(schedule1)
        solution1 = (
            [solution1_r[0], solution1_a[0], solution1_s[0]],
            [[solution1_r[1], solution1_a[1]], solution1_s[1]],
        )

        solution2_r = Recovery.is_recoverable(schedule2)
        if not solution2_r[0]:
            solution2_a = (False, {})
            solution2_s = (False, {})
        else:
            solution2_a = Recovery.avoids_cascading_aborts(schedule2)
        if not solution2_a[0]:
            solution2_s = (False, {})
        else:
            solution2_s = Recovery.is_strict(schedule2)
        solution2 = (
            [solution2_r[0], solution2_a[0], solution2_s[0]],
            [[solution2_r[1], solution2_a[1], solution2_s[1]]],
        )

        solution3_r = Recovery.is_recoverable(schedule3)
        if not solution3_r[0]:
            solution3_a = (False, {})
            solution3_s = (False, {})
        else:
            solution3_a = Recovery.avoids_cascading_aborts(schedule3)
        if not solution3_a[0]:
            solution3_s = (False, {})
        else:
            solution3_s = Recovery.is_strict(schedule3)
        solution3 = (
            [solution3_r[0], solution3_a[0], solution3_s[0]],
            [[solution3_r[1], solution3_a[1], solution3_s[1]]],
        )

        schedule1 = [perform[0], Schedule.parse_string(schedule1)[0]]
        schedule2 = [perform[1], Schedule.parse_string(schedule2)[0]]
        schedule3 = [perform[2], Schedule.parse_string(schedule3)[0]]
        return Task(schedule1, solution1, schedule2, solution2, schedule3, solution3)

    def create_scheduling(self) -> Task:
        schedules = ["SS2PL", "C2PL", "S2PL"]

        for i in range(100):
            schedule1 = generate(3, ["x", "y", "z"], deadlock=False)
            if schedule1[1] == "":
                break
        schedule1 = schedule1[0]
        schedule2 = generate(3, ["x", "y", "z"], deadlock=True)[0]
        for i in range(100):
            schedule3 = generate(3, ["x", "y", "z"], deadlock=False)
            if schedule3[1] == "":
                break
        schedule3 = schedule3[0]
        if (
            Perform_scheduling.perform_SS2PL(schedule1)[1] != ""
            or Perform_scheduling.perform_S2PL(schedule3)[1] != ""
        ):
            raise Exception("A deadlock occured, please try again.")
        solution1 = Perform_scheduling.perform_SS2PL(schedule1)[0]
        solution2 = Perform_scheduling.perform_C2PL(schedule2)
        solution3 = Perform_scheduling.perform_S2PL(schedule3)[0]
        schedule1 = Schedule.parse_string(schedule1)[0]
        schedule2 = Schedule.parse_string(schedule2)[0]
        schedule3 = Schedule.parse_string(schedule3)[0]
        solution1 = Schedule.parse_string(solution1)[0]
        solution2 = Schedule.parse_string(solution2)[0]
        solution3 = Schedule.parse_string(solution3)[0]
        return Task(
            [schedule1, schedules[0]],
            solution1,
            [schedule2, schedules[1]],
            solution2,
            [schedule3, schedules[2]],
            solution3,
        )


def get_tasks():
    tasks = Generate_Tasks()
    return tasks.get_tasks()
